"""Widgets for accessing CaseReader in a Jupyter notebook."""

import re

import numpy as np

from openmdao.recorders.case_reader import CaseReader

# Enable _DEBUG to prevent the debug_output display from clearing itself.
_DEBUG = False

try:
    import matplotlib.pyplot as plt
    import matplotlib as mpl
    from matplotlib import cm
except ImportError:
    mpl = None

try:
    import ipywidgets as ipw
except ImportError:
    ipw = None

try:
    from IPython.display import display
    from IPython import get_ipython
except ImportError:
    get_ipython = None


_func_map = {'None': lambda x: x,
             'ravel': np.ravel,
             'min': np.min,
             'minabs': lambda x: np.min(np.abs(x)),
             'max': np.max,
             'maxabs': lambda x: np.max(np.abs(x)),
             'norm': np.linalg.norm}


def _apply_transform(data, transform):
    """
    Apply a function from _func_map to the data based on the specified transform.

    Parameters
    ----------
    data : np.array
        The data to be transformed.
    transform : str
        The name of the transform to be applied.  This will be looked up in _func_map.
    """
    if transform in _func_map:
        return _func_map[transform](data)
    return data


def _apply_slice(data, s):
    """
    Apply slice s (given as a string) to the data.

    Parameters
    ----------
    data : np.array
        The data to be sliced.
    s : str
        A string representation of the slice, e.g. '[...]', '[0, :]', etc.

    Returns
    -------
    np.array
        The sliced data.
    """
    safe_funcs = {'index_exp': np.index_exp, '__builtins__': None}
    sl = eval(f'index_exp{s}', safe_funcs, None)  # nosec: scope limited
    return data[sl]


def _get_var_meta(cr, case_name, var):
    """
    Return the metadata for the variable of the given name in the given case.

    Parameters
    ----------
    cr : CaseReader
        The case reader housing the data.
    case_name : str
        The case from which the outputs with available metadata is to be returned.
    var : str
        The output whose metadata is desired.

    Returns
    -------
    list
        A dictionary of the metadata for the given output
    """
    case = cr.get_case(case_name)

    case_outputs = case.list_outputs(prom_name=True, units=True, shape=True, val=False,
                                     residuals=False, out_stream=None)

    for _, meta in case_outputs:
        if meta['prom_name'] == var:
            return meta

    case_inputs = case.list_inputs(prom_name=True, units=True, shape=True, val=False,
                                   out_stream=None)

    for _, meta in case_inputs:
        if meta['prom_name'] == var:
            return meta

    raise KeyError(f'No output named {var} found')


def _get_vars(cr, case_names, var_types=None):
    """
    Return a set of variable names whose values are present in at lease one of the given cases.

    Parameters
    ----------
    cr : CaseReader
        The CaseReader housing the data.
    case_names : Iterable of str
        The case_names from which the outputs with available residuals out are to be returned.
    var_types : None or Iterable of str
        The type of variables to be returned.
        Either 'inputs', 'outputs', or None for both inputs and outputs.

    Returns
    -------
    list
        A list of the variables available in at least one of the given cases.
    """
    vars = set()
    for case_name in case_names:
        case = cr.get_case(case_name)
        if var_types is None or 'outputs' in var_types:
            case_outputs = case.list_outputs(prom_name=True, units=False, shape=False,
                                             val=True, residuals=False, out_stream=None)
            vars |= {meta['prom_name'] for abs_path, meta in case_outputs}
        if var_types is None or 'inputs' in var_types:
            case_inputs = case.list_inputs(prom_name=True, units=False, shape=False,
                                           val=True, out_stream=None)
            vars |= {meta['prom_name'] for abs_path, meta in case_inputs}
    return sorted(list(vars))


def _get_resids_vars(cr, case_names):
    """
    Return a set of variable names whose residuals are present in at lease one of the given cases.

    Parameters
    ----------
    cr : CaseReader
        The CaseReader housing the data.
    case_names : Iterable of str
        The case_names from which the outputs with available residuals out are to be returned.

    Returns
    -------
    list
        A list of the variables with available residuals in at least one of the given cases.
    """
    resid_vars = set()
    for case_name in case_names:
        case = cr.get_case(case_name)
        case_outputs = case.list_outputs(prom_name=True, units=True, shape=True, val=False,
                                         residuals=True, out_stream=None)
        resid_vars |= {meta['prom_name'] for abs_path, meta in case_outputs
                       if isinstance(meta['resids'], np.ndarray)}
    return sorted(list(resid_vars))


def _get_opt_vars(cr, case_names, var_type=None):
    """
    Return names of vars that are part of the optimization variables in at least one of the cases.

    Parameters
    ----------
    cr : CaseReader
        The CaseReader housing the data.
    case_names : Iterable of str
        The case_names from which the outputs with available residuals out are to be returned.
    var_type : None or str
        One of 'desvars', 'constraints', 'objectives', or None.

    Returns
    -------
    list
        A list of the variables with available residuals in at least one of the given cases.
    """
    vars = set()
    for case_name in case_names:
        case = cr.get_case(case_name)
        if var_type in ('desvars', None):
            vars |= case.get_design_vars().keys()
        if var_type in ('constraints', None):
            vars |= case.get_constraints().keys()
        if var_type in ('objectives', None):
            vars |= case.get_objectives().keys()
    return sorted(list(vars))


def _get_resids_val(case, prom_name):
    """
    Retrieve residuals associated with the given output in the given case.

    Parameters
    ----------
    case : Case
        The CaseReader Case from which residuals are to be retrieved.
    prom_name : str
        The promoted name of the output whose residuals are to be retrieved.

    Returns
    -------
    np.ndarray
        The residuals of the given output in the given case.

    """
    listed_op = case.list_outputs(prom_name=True, residuals=True, val=False, out_stream=None)
    d = {meta['prom_name']: meta for _, meta in listed_op}
    return d[prom_name]['resids']


def _get_plot_style(i, selected_case_idx, num_cases):
    """
    Return plot attributes based on the index of the selected case and the number of cases.

    Parameters
    ----------
    i : int
        Index of the current case.
    selected_case_idx : int
        Index of the selected case.
    num_cases : int
        Total number of cases being plotted.

    Returns
    -------
    lw : float
        The linewidth attribute for matplotlib line.
    ms : float
        The markersize attribute for matplotlib line.
    s : float
        The s (marker size) attribute for matplotlib scatter.
    alpha : float
        The alpha (transparency) attribute for matplotlib line and scatter.
    """
    if selected_case_idx >= num_cases:
        # Settings when all plots should be displayed "equally".
        alpha = 1.0
        lw = 1
        ms = 2
        s = 10
    else:
        alpha = 0.1 if i != selected_case_idx else 1.0
        lw = 0.5 if i != selected_case_idx else 2
        ms = 2. if i != selected_case_idx else 6.
        s = 10 if i != selected_case_idx else 40
    return lw, ms, s, alpha


class CaseViewer(object):
    """
    Notebook GUI to plot data from a CaseReader.

    Parameters
    ----------
    f : str
        The file from which the cases are to be viewed.

    Attributes
    ----------
    _filename : str
        The filename associated with the record file.
    _case_reader : CaseReader
        The CaseReader instance used to retrieve data from the case file.
    _lines : list of mpl.Line2D
        The line objects on the matplotlib Axes.
    _scatters : list of PathCollection
        The scatter plot objects on the matplotlib Axes.
    _fig : mpl.Figure
        The figure on which the plots are displayed.
    _ax : mpl.Axes
        The matplotlib Axes on which the plots are displayed.
    _cmap : mpl.ColorMap
        The matplotlib color map used to show variation in the plots.
    _scalar_mappable : cm.ScalarMappable
        The object used to provide the extents of the color mapping.
    _colorbar : mpl.Colorbar
        The colorbar shown on the right side of the figure.
    _case_index_str : str
        Constant defining the "Case Index" string used throughout the CaseViewer.
    """

    def __init__(self, f):
        """
        Initialize the case viewer interface.
        """
        if mpl is None:
            raise RuntimeError('CaseViewer requires matplotlib and ipympl')
        if get_ipython is None:
            raise RuntimeError('CaseViewer requires jupyter')
        if ipw is None:
            raise RuntimeError('CaseViewer requires ipywidgets')
        if get_ipython() is None:
            raise RuntimeError('CaseViewer must be run from within a Jupyter notebook.')
        try:
            import ipympl
        except ImportError:
            raise RuntimeError('CaseViewer requires ipympl')

        get_ipython().run_line_magic('matplotlib', 'widget')

        self._case_reader = CaseReader(f) if isinstance(f, str) else f

        self._cmap = cm.viridis

        self._case_index_str = 'Case Index'

        self._filename = self._case_reader._filename

        self._make_gui()

        self._register_callbacks()

        self._fig, self._ax = plt.subplots(1, 1, figsize=(9, 9 / 1.6), tight_layout=True)
        norm = mpl.colors.Normalize(vmin=0, vmax=1)
        self._scalar_mappable = cm.ScalarMappable(norm=norm, cmap=self._cmap)
        self._colorbar = self._fig.colorbar(self._scalar_mappable, label='Case Index', ax=self._ax)
        self._scatters = []
        self._lines = []

        self._update_source_options()
        self._update_case_select_options()
        self._update_var_select_options('x')
        self._update_var_select_options('y')
        self._update_var_info('x')
        self._update_var_info('y')

    def _make_gui(self):
        """
        Define the widgets for the CaseViewer and display them.
        """
        self._widgets = {}

        self._widgets['source_select'] = ipw.Dropdown(description='Source',
                                                      layout=ipw.Layout(width='30%',
                                                                        height='auto'))

        self._widgets['cases_select'] = ipw.SelectMultiple(description='Cases',
                                                           layout=ipw.Layout(width='40%',
                                                                             height='auto'))

        self._widgets['case_select_button'] = ipw.Button(description='Select ' + '\u27F6',
                                                         layout={'width': '100%'})

        self._widgets['case_select_all_button'] = ipw.Button(description='Select All ' + '\u27F9',
                                                             layout={'width': '100%'})

        self._widgets['case_remove_button'] = ipw.Button(description='Remove ' + '\u274c',
                                                         layout={'width': '100%'})

        self._widgets['cases_list'] = ipw.Select(layout=ipw.Layout(width='40%', height='auto'))

        self._widgets['x_filter'] = ipw.Text('', description='X-Axis Filter',
                                             layout=ipw.Layout(width='49%', height='auto'))

        var_types_list = ['outputs', 'inputs', 'optimization', 'desvars',
                          'constraints', 'objectives', 'residuals']

        self._widgets['x_var_type'] = ipw.Dropdown(options=var_types_list,
                                                   description='X Var Type',
                                                   value='outputs',
                                                   layout={'width': '49%'})

        self._widgets['x_select'] = ipw.Select(description='X-Axis',
                                               layout=ipw.Layout(width='auto', height='auto'))

        self._widgets['x_transform_select'] = ipw.Dropdown(options=_func_map.keys(),
                                                           value='None',
                                                           description='X Transform',
                                                           layout=ipw.Layout(width='95%',
                                                                             height='auto'))

        self._widgets['x_slice'] = ipw.Text('[...]', description='X Slice',
                                            layout=ipw.Layout(width='95%', height='auto'))

        self._widgets['x_info'] = ipw.HTML(value='', description='X Shape',
                                           layout={'width': '95%'})

        self._widgets['x_scale'] = ipw.Dropdown(options=['linear', 'log'], value='linear',
                                                description='X Scale', layout={'width': '95%'})

        self._widgets['y_filter'] = ipw.Text('', description='Y-Axis Filter',
                                             layout={'width': '49%', 'height': 'auto'})

        self._widgets['y_var_type'] = ipw.Dropdown(options=var_types_list,
                                                   description='Y Var Type',
                                                   value='outputs',
                                                   layout={'width': '49%'})

        self._widgets['y_select'] = ipw.Select(description='Y-Axis',
                                               layout=ipw.Layout(width='auto', height='auto'))

        self._widgets['y_transform_select'] = ipw.Dropdown(options=_func_map.keys(),
                                                           value='None',
                                                           description='Y Transform',
                                                           layout=ipw.Layout(width='95%',
                                                                             height='auto'))

        self._widgets['y_slice'] = ipw.Text('[...]',
                                            description='Y Slice',
                                            layout=ipw.Layout(width='95%', height='auto'))

        self._widgets['y_info'] = ipw.HTML(value='', description='Y Shape',
                                           layout={'width': '95%'})

        self._widgets['y_scale'] = ipw.Dropdown(options=['linear', 'log'], value='linear',
                                                description='Y Scale',
                                                layout={'width': '95%'})

        self._widgets['case_slider'] = ipw.IntSlider(value=1, min=0, max=1, step=1,
                                                     description='Case #',
                                                     disabled=False, continuous_update=True,
                                                     orientation='horizontal',
                                                     readout=True, readout_format='d',
                                                     layout={'width': '95%'})

        self._widgets['debug_output'] = ipw.Output(description='',
                                                   layout={'border': '0px solid black',
                                                           'width': '95%',
                                                           'height': '400'})

        display(ipw.VBox([self._widgets['source_select'],
                          ipw.HBox([self._widgets['cases_select'],
                                    ipw.VBox([self._widgets['case_select_button'],
                                              self._widgets['case_select_all_button'],
                                              self._widgets['case_remove_button']]),
                                    self._widgets['cases_list']], layout={'width': '95%'}),
                          ipw.HBox([ipw.VBox([ipw.HBox([self._widgets['x_filter'],
                                                        self._widgets['x_var_type']]),
                                              self._widgets['x_select']],
                                             layout={'width': '50%'}),
                                    ipw.VBox([self._widgets['x_info'],
                                              self._widgets['x_slice'],
                                              self._widgets['x_transform_select'],
                                              self._widgets['x_scale']],
                                             layout={'width': '20%'}),
                                    ]),
                          ipw.HBox([ipw.VBox([ipw.HBox([self._widgets['y_filter'],
                                                        self._widgets['y_var_type']]),
                                              self._widgets['y_select']],
                                             layout={'width': '50%'}),
                                    ipw.VBox([self._widgets['y_info'],
                                              self._widgets['y_slice'],
                                              self._widgets['y_transform_select'],
                                              self._widgets['y_scale']],
                                             layout={'width': '20%'})
                                    ]),
                          self._widgets['case_slider'],
                          self._widgets['debug_output']]))

    def _update_source_options(self):
        """
        Update the contents of the source selection dropdown menu.
        """
        sources = self._case_reader.list_sources(out_stream=None)
        self._widgets['source_select'].options = sources
        self._widgets['source_select'].value = sources[0]

    def _update_var_info(self, axis):
        """
        Update the variable info displayed.

        Parameters
        ----------
        axis : str
            'x' or 'y' - case insensitive.
        """
        if axis.lower() not in ('x', 'y'):
            raise ValueError(f'Unknown axis: {axis}')

        src = self._widgets['source_select'].value
        cases = self._widgets['cases_list'].options

        if not cases:
            self._widgets[f'{axis}_info'].value = 'N/A'
            return

        var = self._widgets[f'{axis}_select'].value

        if var == self._case_index_str:
            shape = (len(cases),)
        elif var is None:
            shape = 'N/A'
        else:
            meta = _get_var_meta(self._case_reader, cases[0], var)
            shape = meta['shape']

        self._widgets[f'{axis}_info'].value = f'{shape}'

    def _update_case_select_options(self):
        """
        Update the available cases listed in the source_select widget.
        """
        src = self._widgets['source_select'].value
        avialable_cases = self._case_reader.list_cases(source=src, recurse=False, out_stream=None)
        self._widgets['cases_select'].options = avialable_cases
        self._update_case_slider()

    def _update_var_select_options(self, axis):
        """
        Update the variables available for plotting.

        Parameters
        ----------
        axis : str
            'x' or 'y' - case insensitive.
        """
        with self._widgets['debug_output']:
            if axis.lower() not in ('x', 'y'):
                raise ValueError(f'Unknown axis: {axis}')

            src = self._widgets['source_select'].value
            cases = self._widgets['cases_list'].options

            if not cases:
                self._widgets[f'{axis}_select'].options = []
                self._widgets[f'{axis}_info'].value = 'N/A'
                return

            w_var_select = self._widgets[f'{axis}_select']
            var_filter = self._widgets[f'{axis}_filter'].value
            var_select = w_var_select.value
            var_type = self._widgets[f'{axis}_var_type'].value

            if var_type == 'optimization':
                vars = _get_opt_vars(self._case_reader, cases)
            elif var_type in ('desvars', 'constraints', 'objectives'):
                vars = _get_opt_vars(self._case_reader, cases, var_type=var_type)
            elif var_type == 'residuals':
                vars = _get_resids_vars(self._case_reader, cases)
            else:
                vars = _get_vars(self._case_reader, cases, var_types=var_type)

            # We have a list of available vars, now filter it.
            r = re.compile(var_filter)
            filtered_list = list(filter(r.search, vars))

            w_var_select.options = [self._case_index_str] + filtered_list \
                if axis == 'x' else filtered_list

            self._update_var_info(axis)

    def _update_case_slider(self):
        """
        Update the extents of the case slider.
        """
        n = len(self._widgets['cases_list'].options)
        self._widgets['case_slider'].max = n
        self._widgets['case_slider'].value = n

    def _register_callbacks(self):
        """
        Register callback functions with the widgets.
        """
        self._widgets['debug_output'].clear_output()
        with self._widgets['debug_output']:

            self._widgets['source_select'].observe(self._callback_select_source)
            self._widgets['case_select_button'].on_click(self._callback_select_case)
            self._widgets['case_select_all_button'].on_click(self._callback_select_all_cases)
            self._widgets['case_remove_button'].on_click(self._callback_remove_case)

            self._widgets['cases_list'].observe(self._callback_case_list_select)

            self._widgets['x_filter'].observe(self._callback_filter_vars, 'value')
            self._widgets['y_filter'].observe(self._callback_filter_vars, 'value')

            self._widgets['x_var_type'].observe(self._callback_filter_vars, 'value')
            self._widgets['y_var_type'].observe(self._callback_filter_vars, 'value')

            self._widgets['x_select'].observe(self._callback_select_var, 'value')
            self._widgets['y_select'].observe(self._callback_select_var, 'value')

            self._widgets['x_slice'].observe(self._callback_change_slice, 'value')
            self._widgets['y_slice'].observe(self._callback_change_slice, 'value')

            self._widgets['x_scale'].observe(self._callback_change_scale, 'value')
            self._widgets['y_scale'].observe(self._callback_change_scale, 'value')

            self._widgets['x_transform_select'].observe(self._callback_select_transform, 'value')
            self._widgets['y_transform_select'].observe(self._callback_select_transform, 'value')

            self._widgets['case_slider'].observe(self._callback_case_slider, 'value')

    def _callback_select_source(self, *args):
        """
        Repopulate cases_select with cases from the chosen source.

        Parameters
        ----------
        args : tuple
            The information passed by the widget upon callback.
        """
        self._widgets['debug_output'].clear_output()
        with self._widgets['debug_output']:
            self._update_case_select_options()

    def _callback_select_case(self, *args):
        """
        Add the selected case(s) to the chosen cases list.

        Parameters
        ----------
        args : tuple
            The information passed by the widget upon callback.
        """
        self._widgets['debug_output'].clear_output()
        with self._widgets['debug_output']:
            clw = self._widgets['cases_list']
            current = clw.options
            new = self._widgets['cases_select'].value

            def _numeric_sorter(case_name):
                return case_name.split('|')[0], int(case_name.split('|')[-1])

            self._widgets['cases_list'].options = sorted(list(set(current + new)),
                                                         key=_numeric_sorter)
            self._update_case_slider()
            self._update_var_select_options('x')
            self._update_var_select_options('y')
            self._update_plot()

    def _callback_case_list_select(self, *args):
        """
        Update the plot when a different case is selected in the cases list.

        Parameters
        ----------
        args : tuple
            The information passed by the widget upon callback.
        """
        if not _DEBUG:
            self._widgets['debug_output'].clear_output()
        with self._widgets['debug_output']:
            self._update_plot()

    def _callback_case_slider(self, *args):
        """
        Update the plot when the case slider is changed.

        Parameters
        ----------
        args : tuple
            The information passed by the widget upon callback.
        """
        if not _DEBUG:
            self._widgets['debug_output'].clear_output()
        with self._widgets['debug_output']:
            selected_case_idx = self._widgets['case_slider'].value
            cases = self._widgets['cases_list'].options

            if not _DEBUG:
                self._widgets['debug_output'].clear_output()
            with self._widgets['debug_output']:
                for i, case_name in enumerate(cases):
                    lw, ms, s, alpha = _get_plot_style(i, selected_case_idx, len(cases))
                    if i < len(self._scatters):
                        scat = self._scatters[i]
                        scat.set_sizes(s * np.ones_like(scat.get_sizes()))
                        scat.set_alpha(alpha)
                    if i < len(self._lines):
                        line = self._lines[i]
                        for li in line:
                            li.set_linewidth(lw)
                            li.set_markersize(ms)
                            li.set_alpha(alpha)

    def _callback_select_all_cases(self, *args):
        """
        Add all available cases to the cases list.

        Parameters
        ----------
        args : tuple
            The information passed by the widget upon callback.
        """
        if not _DEBUG:
            self._widgets['debug_output'].clear_output()
        with self._widgets['debug_output']:
            clw = self._widgets['cases_list']
            current = clw.options
            new = self._widgets['cases_select'].options

            def _numeric_sorter(case_name):
                return case_name.split('|')[0], int(case_name.split('|')[-1])

            self._widgets['cases_list'].options = sorted(list(set(current + new)),
                                                         key=_numeric_sorter)
            self._update_case_slider()
            self._update_var_select_options('x')
            self._update_var_select_options('y')
            self._update_plot()

    def _callback_remove_case(self, *args):
        """
        Remove the selected case from the chosen cases list widget.

        Parameters
        ----------
        args : tuple
            The information passed by the widget upon callback.
        """
        if not _DEBUG:
            self._widgets['debug_output'].clear_output()
        with self._widgets['debug_output']:
            clw = self._widgets['cases_list']
            new_list = list(clw.options)
            if clw.value in new_list:
                new_list.remove(clw.value)
            clw.options = new_list
            self._update_var_select_options('x')
            self._update_var_select_options('y')
            self._update_plot()

    def _callback_filter_vars(self, *args):
        """
        Update the plot and the available variables when the filtering criteria is changed.

        Parameters
        ----------
        args : tuple
            The information passed by the widget upon callback.
        """
        event = args[0]
        if not _DEBUG:
            self._widgets['debug_output'].clear_output()
        with self._widgets['debug_output']:
            w = event['owner']
            axis = 'x' if w is self._widgets['x_filter'] \
                or w is self._widgets['x_var_type'] else 'y'
            self._update_var_select_options(axis)
            self._update_plot()

    def _callback_select_var(self, *args):
        """
        Update the variable info and the plot when a new variable is chosen.

        Parameters
        ----------
        args : tuple
            The information passed by the widget upon callback.
        """
        event = args[0]
        if not _DEBUG:
            self._widgets['debug_output'].clear_output()
        with self._widgets['debug_output']:
            w = event['owner']
            s = w.value
            axis = 'x' if w is self._widgets['x_select'] else 'y'

            self._update_var_info(axis)

            if s is None:
                self._ax.clear()
            else:
                self._update_plot()

    def _callback_change_slice(self, *args):
        """
        Update the plot when a new, valid slice is provided.

        Parameters
        ----------
        args : tuple
            The information passed by the widget upon callback.
        """
        event = args[0]
        if not _DEBUG:
            self._widgets['debug_output'].clear_output()
        with self._widgets['debug_output']:
            w = event['owner']
            s = w.value

            if s.startswith('[') and s.endswith(']'):
                self._update_plot()

    def _callback_change_scale(self, *args):
        """
        Update the plot when the x or y axis scale is changed.

        Parameters
        ----------
        args : tuple
            The information passed by the widget upon callback.
        """
        if not _DEBUG:
            self._widgets['debug_output'].clear_output()
        with self._widgets['debug_output']:
            event = args[0]
            w = event['owner']
            if w is self._widgets['x_scale']:
                self._ax.set_xscale(w.value)
            else:
                self._ax.set_yscale(w.value)

    def _callback_select_transform(self, *args):
        """
        Update the plot when a new transformation is choen for the x or y variable.

        Parameters
        ----------
        args : tuple
            The information passed by the widget upon callback.
        """
        self._update_plot()

    def _redraw_plot(self):
        """
        Update the plot area by plotting one variable vs another over one or more cases.
        """
        x_min = y_min = 1E16
        x_max = y_max = -1E16

        cases = self._widgets['cases_list'].options

        x_slice = self._widgets['x_slice'].value
        y_slice = self._widgets['y_slice'].value

        x_transform = self._widgets['x_transform_select'].value
        y_transform = self._widgets['y_transform_select'].value

        x_var = self._widgets['x_select'].value
        y_var = self._widgets['y_select'].value

        x_var_type = self._widgets['x_var_type'].value
        y_var_type = self._widgets['y_var_type'].value

        selected_case_idx = self._widgets['case_slider'].value

        max_size = 0

        if not _DEBUG:
            self._widgets['debug_output'].clear_output()
        with self._widgets['debug_output']:
            for i, case_name in enumerate(cases):
                lw, ms, s, alpha = _get_plot_style(i, selected_case_idx, len(cases))
                case = self._case_reader.get_case(case_name)

                if y_var_type == 'residuals':
                    y_val = _get_resids_val(case, y_var)
                else:
                    y_val = case.get_val(y_var)

                try:
                    y_val = _apply_slice(y_val, y_slice)
                except ValueError:
                    if _DEBUG:
                        print(f'Error while applying Y slice: {y_slice}')
                    continue
                y_val = _apply_transform(y_val, y_transform)

                if x_var != self._case_index_str:
                    if x_var_type == 'residuals':
                        x_val = _get_resids_val(case, x_var)
                    else:
                        x_val = self._case_reader.get_case(case_name).get_val(x_var)
                else:
                    x_val = i * np.ones(y_val.size)

                try:
                    x_val = _apply_slice(x_val, x_slice)
                except ValueError:
                    if _DEBUG:
                        print(f'Error while applying X slice: {x_slice}')
                    continue
                x_val = _apply_transform(x_val, x_transform)

                if x_val is None or y_val is None:
                    continue

                if x_val.shape[0] != y_val.shape[0]:
                    print(f'Incompatible shapes: x.shape = {x_val.shape} '
                          f' y.shape = {y_val.shape}.')
                    print('Size along first axis must agree.')
                    return

                max_size = max(max_size, x_val.size)

                x_min = min(x_min, np.min(x_val))
                x_max = max(x_max, np.max(x_val))
                y_min = min(y_min, np.min(y_val))
                y_max = max(y_max, np.max(y_val))

                if x_var == self._case_index_str:
                    s = self._ax.scatter(x_val, y_val, c=np.arange(x_val.size), s=s,
                                         cmap=self._cmap, alpha=alpha)
                    self._scatters.append(s)
                else:
                    line = self._ax.plot(x_val, y_val,
                                         c=self._cmap(float(i) / len(cases)),
                                         mfc=self._cmap(float(i) / len(cases)),
                                         mec='k',
                                         marker='o',
                                         linestyle='-',
                                         linewidth=lw,
                                         markersize=ms,
                                         alpha=alpha)
                    self._lines.append(line)

                self._fig.canvas.flush_events()

            x_margin = (x_max - x_min) * 0.05
            x_margin = 0.1 if x_margin < 1.0E-16 else x_margin
            y_margin = (y_max - y_min) * 0.05
            y_margin = 0.1 if y_margin < 1.0E-16 else y_margin

            bad_x_bounds = np.any(x.isinf() or x.isnan() for x in [x_min, x_max, x_margin])
            bad_y_bounds = np.any(x.isinf() or x.isnan() for x in [x_min, x_max, x_margin])

            if not bad_x_bounds:
                self._ax.set_xlim(x_min - x_margin, x_max + x_margin)
            if not bad_y_bounds:
                self._ax.set_ylim(y_min - y_margin, y_max + y_margin)

            # Add the colorbar.
            # Color shows the index of each point in its vector if the x-axis is Case Index,
            # otherwise it shows the case index.
            if x_var == self._case_index_str:
                vmax = max_size
                cbar_label = 'Array Index'
            else:
                vmax = len(cases)
                cbar_label = self._case_index_str
            self._scalar_mappable.set_clim(0, vmax)
            self._colorbar.set_label(cbar_label)

    def _update_plot(self):
        """
        Update the plot based on the contents of the widgets.
        """
        with self._widgets['debug_output']:
            cr = self._case_reader
            src = self._widgets['source_select'].value
            cases = self._widgets['cases_list'].options
            x_var = self._widgets['x_select'].value
            y_var = self._widgets['y_select'].value
            x_slice = '' if self._widgets['x_slice'].value == '[...]' \
                else self._widgets['x_slice'].value
            y_slice = '' if self._widgets['y_slice'].value == '[...]' \
                else self._widgets['y_slice'].value
            x_transform = self._widgets['x_transform_select'].value
            y_transform = self._widgets['y_transform_select'].value
            x_var_type = self._widgets['x_var_type'].value
            y_var_type = self._widgets['y_var_type'].value

            try:
                self._ax.clear()
                self._scatters.clear()
                self._lines.clear()
            except AttributeError:
                return

            if not cases or not x_var or not y_var:
                print('Nothing to plot')
                return

            x_units = 'None' if x_var == self._case_index_str \
                else _get_var_meta(cr, cases[0], x_var)['units']
            y_units = _get_var_meta(cr, cases[0], y_var)['units']

            self._redraw_plot()

            x_label = rf'{x_var}{x_slice}'
            y_label = rf'{y_var}{y_slice}'

            if x_var_type == 'residuals':
                x_label = rf'$\mathcal{{R}}$({x_label})'

            if y_var_type == 'residuals':
                y_label = rf'$\mathcal{{R}}$({y_label})'

            if x_transform != 'None':
                x_label = f'{x_transform}({x_label})'

            if y_transform != 'None':
                y_label = f'{y_transform}({y_label})'

            self._ax.set_xlabel(f'{x_label}\n({x_units})')
            self._ax.set_ylabel(f'{y_label}\n({y_units})')
            self._ax.grid(True)

            self._ax.set_xscale(self._widgets['x_scale'].value)
            self._ax.set_yscale(self._widgets['y_scale'].value)



""" Unit test for the solver printing behavior. """

import os
from io import StringIO
import re
import sys
import unittest

import numpy as np

import openmdao.api as om
from openmdao.test_suite.components.double_sellar import SubSellar
from openmdao.test_suite.components.sellar import SellarDerivatives

from openmdao.utils.general_utils import run_model
from openmdao.utils.mpi import MPI


try:
    from openmdao.vectors.petsc_vector import PETScVector
except ImportError:
    PETScVector = None


class TestSolverPrint(unittest.TestCase):

    def test_feature_iprint_neg1(self):

        prob = om.Problem()
        prob.model = SellarDerivatives()

        prob.setup()

        newton = prob.model.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
        scipy = prob.model.linear_solver = om.ScipyKrylov()

        newton.options['maxiter'] = 2

        # use a real bad initial guess
        prob.set_val('y1', 10000)
        prob.set_val('y2', -26)

        newton.options['iprint'] = -1
        scipy.options['iprint'] = -1
        prob.run_model()

    def test_iprint_neg1(self):
        old_stdout = sys.stdout
        sys.stdout = str_out = StringIO()

        try:
            self.test_feature_iprint_neg1()
        finally:
            sys.stdout = old_stdout

        # Verify output
        self.assertEqual(str_out.getvalue().strip(), "")

    def test_feature_iprint_0(self):

        prob = om.Problem()
        prob.model = SellarDerivatives()

        prob.setup()

        newton = prob.model.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
        scipy = prob.model.linear_solver = om.ScipyKrylov()

        newton.options['maxiter'] = 1

        prob.set_val('y1', 10000)
        prob.set_val('y2', -26)

        newton.options['iprint'] = 0
        scipy.options['iprint'] = 0

        prob.run_model()

    def test_iprint_0(self):
        old_stdout = sys.stdout
        sys.stdout = str_out = StringIO()

        try:
            self.test_feature_iprint_0()
        finally:
            sys.stdout = old_stdout

        # Verify output
        self.assertEqual(str_out.getvalue().strip(),
                         "NL: NewtonSolver 'NL: Newton' on system '' failed to converge in "
                         "1 iterations.")

    def test_feature_iprint_1(self, stdout=None):

        self.prob = prob = om.Problem()
        prob.model = SellarDerivatives()

        prob.setup()

        newton = prob.model.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
        scipy = prob.model.linear_solver = om.ScipyKrylov()

        newton.options['maxiter'] = 20

        prob.set_val('y1', 10000)
        prob.set_val('y2', -26)

        newton.options['iprint'] = 1
        scipy.options['iprint'] = 0
        prob.run_model()

    def test_iprint_1(self):
        old_stdout = sys.stdout
        sys.stdout = str_out = StringIO()

        try:
            self.test_feature_iprint_1()
        finally:
            sys.stdout = old_stdout

        iters = self.prob.model.nonlinear_solver._iter_count

        # Verify output
        self.assertEqual(str_out.getvalue().strip(),
                         f"NL: Newton Converged in {iters} iterations")

    def test_feature_iprint_2(self):

        prob = om.Problem()
        prob.model = SellarDerivatives()

        prob.setup()

        newton = prob.model.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
        scipy = prob.model.linear_solver = om.ScipyKrylov()

        newton.options['maxiter'] = 20

        prob.set_val('y1', 10000)
        prob.set_val('y2', -26)

        newton.options['iprint'] = 2
        scipy.options['iprint'] = 1
        prob.run_model()

    def test_iprint_2(self):
        old_stdout = sys.stdout
        sys.stdout = str_out = StringIO()

        try:
            self.test_feature_iprint_2()
        finally:
            sys.stdout = old_stdout

        # Verify output
        lines = str_out.getvalue().strip().split('\n')
        for line in lines:
            self.assertEqual(line[:3], 'NL:')

    def test_hierarchy_iprint(self):

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('pz', om.IndepVarComp('z', np.array([5.0, 2.0])))

        sub1 = model.add_subsystem('sub1', om.Group())
        sub2 = sub1.add_subsystem('sub2', om.Group())
        g1 = sub2.add_subsystem('g1', SubSellar())
        g2 = model.add_subsystem('g2', SubSellar())

        model.connect('pz.z', 'sub1.sub2.g1.z')
        model.connect('sub1.sub2.g1.y2', 'g2.x')
        model.connect('g2.y2', 'sub1.sub2.g1.x')

        model.nonlinear_solver = om.NewtonSolver()
        model.linear_solver = om.ScipyKrylov()
        model.nonlinear_solver.options['solve_subsystems'] = True
        model.nonlinear_solver.options['max_sub_solves'] = 0

        g1.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
        g1.linear_solver = om.LinearBlockGS()

        g2.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
        g2.linear_solver = om.ScipyKrylov()
        g2.linear_solver.precon = om.LinearBlockGS()
        g2.linear_solver.precon.options['maxiter'] = 2

        prob.set_solver_print(level=2)

        prob.setup()

        output = run_model(prob)

        # Check that certain things show up in our outputs
        self.assertGreaterEqual(output.count('sub1.sub2.g1'), 1)
        self.assertGreaterEqual(output.count('g2'), 1)
        self.assertGreaterEqual(output.count('NL: Newton'), 2)
        self.assertGreaterEqual(output.count('LN: SCIPY'), 2)
        self.assertGreaterEqual(output.count('precon: LNBGS'), 2)

    def test_hierarchy_iprint2(self):

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('pz', om.IndepVarComp('z', np.array([5.0, 2.0])))

        sub1 = model.add_subsystem('sub1', om.Group())
        sub2 = sub1.add_subsystem('sub2', om.Group())
        g1 = sub2.add_subsystem('g1', SubSellar())
        g2 = model.add_subsystem('g2', SubSellar())

        model.connect('pz.z', 'sub1.sub2.g1.z')
        model.connect('sub1.sub2.g1.y2', 'g2.x')
        model.connect('g2.y2', 'sub1.sub2.g1.x')

        model.nonlinear_solver = om.NonlinearBlockGS()
        g1.nonlinear_solver = om.NonlinearBlockGS()
        g2.nonlinear_solver = om.NonlinearBlockGS()

        prob.set_solver_print(level=2)

        prob.setup()

        output = run_model(prob)

        # Check that certain things show up in our outputs
        self.assertGreaterEqual(output.count('sub1.sub2.g1'), 2)
        self.assertGreaterEqual(output.count('g2'), 2)
        self.assertGreaterEqual(output.count('NL: NLBGS'), 2)

    def test_hierarchy_iprint3(self):

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('pz', om.IndepVarComp('z', np.array([5.0, 2.0])))

        sub1 = model.add_subsystem('sub1', om.Group())
        sub2 = sub1.add_subsystem('sub2', om.Group())
        g1 = sub2.add_subsystem('g1', SubSellar())
        g2 = model.add_subsystem('g2', SubSellar())

        model.connect('pz.z', 'sub1.sub2.g1.z')
        model.connect('sub1.sub2.g1.y2', 'g2.x')
        model.connect('g2.y2', 'sub1.sub2.g1.x')

        model.nonlinear_solver = om.NonlinearBlockJac()
        sub1.nonlinear_solver = om.NonlinearBlockJac()
        sub2.nonlinear_solver = om.NonlinearBlockJac()
        g1.nonlinear_solver = om.NonlinearBlockJac()
        g2.nonlinear_solver = om.NonlinearBlockJac()

        prob.set_solver_print(level=2)

        prob.setup()

        output = run_model(prob)

        # Check that certain things show up in our outputs
        self.assertGreaterEqual(output.count('sub1'), 2)
        self.assertGreaterEqual(output.count('sub1.sub2'), 2)
        self.assertGreaterEqual(output.count('sub1.sub2.g1'), 2)
        self.assertGreaterEqual(output.count('g2'), 2)
        self.assertGreaterEqual(output.count('NL: NLBJ'), 2)

    def test_feature_set_solver_print1(self):

        prob = om.Problem()
        model = prob.model

        sub1 = model.add_subsystem('sub1', om.Group(), promotes_inputs=['z'])
        sub2 = sub1.add_subsystem('sub2', om.Group(), promotes_inputs=['z'])
        g1 = sub2.add_subsystem('g1', SubSellar(), promotes_inputs=['z'])
        g2 = model.add_subsystem('g2', SubSellar())

        model.connect('sub1.sub2.g1.y2', 'g2.x')
        model.connect('g2.y2', 'sub1.sub2.g1.x')

        model.nonlinear_solver = om.NewtonSolver()
        model.linear_solver = om.ScipyKrylov()
        model.nonlinear_solver.options['solve_subsystems'] = True
        model.nonlinear_solver.options['max_sub_solves'] = 0

        g1.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
        g1.linear_solver = om.LinearBlockGS()

        g2.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
        g2.linear_solver = om.ScipyKrylov()
        g2.linear_solver.precon = om.LinearBlockGS()
        g2.linear_solver.precon.options['maxiter'] = 2

        prob.set_solver_print(level=2)

        prob.setup()
        prob.set_val('z', np.array([5.0, 2.0]))
        prob.run_model()

    def test_set_solver_print1(self):
        old_stdout = sys.stdout
        sys.stdout = str_out = StringIO()

        try:
            self.test_feature_set_solver_print1()
        finally:
            sys.stdout = old_stdout

        # Check that certain things show up in our outputs
        output = str_out.getvalue()
        self.assertGreaterEqual(output.count('sub1.sub2.g1'), 2)
        self.assertGreaterEqual(output.count('g2'), 2)
        self.assertGreaterEqual(output.count('NL: Newton'), 2)
        self.assertGreaterEqual(output.count('LN: LNBGS'), 2)
        self.assertGreaterEqual(output.count('LN: SCIPY'), 2)
        self.assertGreaterEqual(output.count('LS: BCHK'), 2)
        self.assertGreaterEqual(output.count('precon: LNBGS'), 2)


    def test_feature_set_solver_print2(self):

        prob = om.Problem()
        model = prob.model

        sub1 = model.add_subsystem('sub1', om.Group(), promotes_inputs=['z'])
        sub2 = sub1.add_subsystem('sub2', om.Group(), promotes_inputs=['z'])
        g1 = sub2.add_subsystem('g1', SubSellar(), promotes_inputs=['z'])
        g2 = model.add_subsystem('g2', SubSellar())

        model.connect('sub1.sub2.g1.y2', 'g2.x')
        model.connect('g2.y2', 'sub1.sub2.g1.x')

        model.nonlinear_solver = om.NewtonSolver()
        model.linear_solver = om.ScipyKrylov()
        model.nonlinear_solver.options['solve_subsystems'] = True
        model.nonlinear_solver.options['max_sub_solves'] = 0

        g1.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
        g1.linear_solver = om.LinearBlockGS()

        g2.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
        g2.linear_solver = om.ScipyKrylov()
        g2.linear_solver.precon = om.LinearBlockGS()
        g2.linear_solver.precon.options['maxiter'] = 2

        prob.set_solver_print(level=2)
        prob.set_solver_print(level=-1, type_='LN')

        prob.setup()
        prob.set_val('z', np.array([5.0, 2.0]))
        prob.run_model()

    def test_set_solver_print2(self):
        old_stdout = sys.stdout
        sys.stdout = str_out = StringIO()

        try:
            self.test_feature_set_solver_print2()
        finally:
            sys.stdout = old_stdout

        # Check that certain things show up in our outputs
        output = str_out.getvalue()
        self.assertGreaterEqual(output.count('sub1.sub2.g1'), 1)
        self.assertGreaterEqual(output.count('g2'), 1)
        self.assertGreaterEqual(output.count('NL: Newton'), 2)
        self.assertGreaterEqual(output.count('LS: BCHK'), 2)

    def test_feature_set_solver_print3(self):

        prob = om.Problem()
        model = prob.model

        sub1 = model.add_subsystem('sub1', om.Group(), promotes_inputs=['z'])
        sub2 = sub1.add_subsystem('sub2', om.Group(), promotes_inputs=['z'])
        g1 = sub2.add_subsystem('g1', SubSellar(), promotes_inputs=['z'])
        g2 = model.add_subsystem('g2', SubSellar())

        model.connect('sub1.sub2.g1.y2', 'g2.x')
        model.connect('g2.y2', 'sub1.sub2.g1.x')

        model.nonlinear_solver = om.NewtonSolver()
        model.linear_solver = om.ScipyKrylov()
        model.nonlinear_solver.options['solve_subsystems'] = True
        model.nonlinear_solver.options['max_sub_solves'] = 0

        g1.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
        g1.linear_solver = om.LinearBlockGS()

        g2.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
        g2.linear_solver = om.ScipyKrylov()
        g2.linear_solver.precon = om.LinearBlockGS()
        g2.linear_solver.precon.options['maxiter'] = 2

        prob.set_solver_print(level=0)
        prob.set_solver_print(level=2, depth=2)

        prob.setup()
        prob.set_val('z', np.array([5.0, 2.0]))
        prob.run_model()

    def test_set_solver_print3(self):
        old_stdout = sys.stdout
        sys.stdout = str_out = StringIO()

        try:
            self.test_feature_set_solver_print3()
        finally:
            sys.stdout = old_stdout

        # Check that certain things show up in our outputs
        output = str_out.getvalue()
        self.assertGreaterEqual(output.count('sub1.sub2.g1'), 1)
        self.assertGreaterEqual(output.count('g2'), 1)
        self.assertGreaterEqual(output.count('NL: Newton'), 2)
        self.assertGreaterEqual(output.count('precon: LNBGS'), 2)
        self.assertGreaterEqual(output.count('LS: BCHK'), 2)
        self.assertGreaterEqual(output.count('LN: SCIPY'), 2)

    def test_feature_set_solver_print4(self):

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('pz', om.IndepVarComp('z', np.array([5.0, 2.0])))

        sub1 = model.add_subsystem('sub1', om.Group())
        sub2 = sub1.add_subsystem('sub2', om.Group())
        g1 = sub2.add_subsystem('g1', SubSellar())
        g2 = model.add_subsystem('g2', SubSellar())

        model.connect('sub1.sub2.g1.y2', 'g2.x')
        model.connect('g2.y2', 'sub1.sub2.g1.x')

        model.nonlinear_solver = om.NewtonSolver()
        model.linear_solver = om.ScipyKrylov()
        model.nonlinear_solver.options['solve_subsystems'] = True
        model.nonlinear_solver.options['max_sub_solves'] = 0

        g1.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
        g1.linear_solver = om.LinearBlockGS()

        g2.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
        g2.linear_solver = om.ScipyKrylov()
        g2.linear_solver.precon = om.LinearBlockGS()
        g2.linear_solver.precon.options['maxiter'] = 2

        prob.set_solver_print(level=-1, type_='all')
        g2.set_solver_print(level=2, type_='NL')

        prob.setup()
        prob.run_model()

    def test_set_solver_print4(self):
        old_stdout = sys.stdout
        sys.stdout = str_out = StringIO()

        try:
            self.test_feature_set_solver_print4()
        finally:
            sys.stdout = old_stdout

        # Check that certain things show up in our outputs
        output = str_out.getvalue()
        self.assertEqual(output.count('sub1.sub2.g1'), 0)
        self.assertGreaterEqual(output.count('g2'), 1)
        self.assertGreaterEqual(output.count('NL: Newton'), 2)
        self.assertEqual(output.count('precon: LNBGS'), 0)
        self.assertGreaterEqual(output.count('LS: BCHK'), 1)
        self.assertEqual(output.count('LN: SCIPY'), 0)


@unittest.skipUnless(MPI and PETScVector, "MPI and PETSc are required.")
class MPITests(unittest.TestCase):

    N_PROCS = 2

    def test_hierarchy_iprint(self):
        prob = om.Problem()
        model = prob.model

        model.add_subsystem('pz', om.IndepVarComp('z', np.array([5.0, 2.0])))

        sub1 = model.add_subsystem('sub1', om.Group())
        sub2 = sub1.add_subsystem('sub2', om.Group())
        g1 = sub2.add_subsystem('g1', SubSellar())
        g2 = model.add_subsystem('g2', SubSellar())

        model.connect('pz.z', 'sub1.sub2.g1.z')
        model.connect('sub1.sub2.g1.y2', 'g2.x')
        model.connect('g2.y2', 'sub1.sub2.g1.x')

        model.nonlinear_solver = om.NewtonSolver()
        model.linear_solver = om.LinearBlockGS()
        model.nonlinear_solver.options['solve_subsystems'] = True
        model.nonlinear_solver.options['max_sub_solves'] = 0

        g1.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
        g1.linear_solver = om.LinearBlockGS()

        g2.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
        g2.linear_solver = om.PETScKrylov()
        g2.linear_solver.precon = om.LinearBlockGS()
        g2.linear_solver.precon.options['maxiter'] = 2

        prob.set_solver_print(level=2)

        prob.setup()

        # Conclude setup but don't run model.
        prob.final_setup()

        # if USE_PROC_FILES is not set, solver convergence messages
        # should only appear on proc 0
        output = run_model(prob)
        if model.comm.rank == 0 or os.environ.get('USE_PROC_FILES'):
            self.assertTrue(output.count('\nNL: Newton Converged') == 1)
        else:
            self.assertTrue(output.count('\nNL: Newton Converged') == 0)


if __name__ == "__main__":
    unittest.main()



""" Test the ExternalCodeComp. """
import os
import sys
import shutil
import tempfile
import unittest

from scipy.optimize import fsolve

import openmdao.api as om
from openmdao.components.external_code_comp import STDOUT

from openmdao.utils.assert_utils import assert_near_equal, assert_warning

DIRECTORY = os.path.dirname((os.path.abspath(__file__)))


# These next three functions are used by test_simple_external_code_implicit_comp_with_solver
def area_ratio_explicit(Mach):
    """isentropic relationship between area ratio and Mach number"""
    gamma = 1.4
    gamma_p_1 = gamma + 1
    gamma_m_1 = gamma - 1
    exponent = gamma_p_1 / (2 * gamma_m_1)

    return (gamma_p_1 / 2.) ** -exponent * ((1 + gamma_m_1 / 2. * Mach ** 2) ** exponent) / Mach


def mach_residual(Mach, area_ratio_target):
    return area_ratio_target - area_ratio_explicit(Mach)


def mach_solve(area_ratio, super_sonic=False):
    if super_sonic:
        initial_guess = 4
    else:
        initial_guess = .1

    mach = fsolve(func=mach_residual, x0=initial_guess, args=(area_ratio,))[0]

    return mach


class TestExternalCodeComp(unittest.TestCase):

    def setUp(self):
        self.startdir = os.getcwd()
        self.tempdir = tempfile.mkdtemp(prefix='test_extcode-')
        os.chdir(self.tempdir)
        shutil.copy(os.path.join(DIRECTORY, 'extcode_example.py'),
                    os.path.join(self.tempdir, 'extcode_example.py'))

        self.prob = om.Problem()

        self.extcode = self.prob.model.add_subsystem('extcode', om.ExternalCodeComp())

    def tearDown(self):
        os.chdir(self.startdir)
        try:
            shutil.rmtree(self.tempdir)
        except OSError:
            pass

    def test_normal(self):
        self.extcode.options['command'] = [
            sys.executable, 'extcode_example.py', 'extcode.out'
        ]

        self.extcode.options['external_input_files'] = ['extcode_example.py']
        self.extcode.options['external_output_files'] = ['extcode.out']

        self.prob.setup(check=True)
        self.prob.run_model()

        with open('extcode.out', 'r') as f:
            self.assertEqual(f.read(), 'test data\n')

    @unittest.skipUnless(sys.platform == 'win32', 'Windows-specific test.')
    def test_normal_bat(self):
        batch_script = '\n'.join([
            "@echo off",
            "rem usage: extcode.bat output_filename",
            "rem",
            "rem Just write 'test data' to the specified output file",
            "",
            "set DATA=test data",
            "set OUT_FILE=%1",
            "",
            "echo %DATA%>>%OUT_FILE%"
        ])
        with open('extcode.bat', 'w') as f:
            f.write(batch_script)

        self.extcode.options['command'] = [
            'extcode.bat', 'extcode.out'
        ]

        self.extcode.options['external_input_files'] = ['extcode.bat']
        self.extcode.options['external_output_files'] = ['extcode.out']

        self.prob.setup(check=True)
        self.prob.run_model()

        with open('extcode.out', 'r') as f:
            self.assertEqual(f.read(), 'test data\n')

    def test_timeout_raise(self):
        self.extcode.options['command'] = [
            sys.executable, 'extcode_example.py', 'extcode.out', '--delay', '3'
        ]
        self.extcode.options['timeout'] = 1.0

        self.extcode.options['external_input_files'] = ['extcode_example.py']

        self.prob.setup(check=True)
        try:
            self.prob.run_model()
        except om.AnalysisError as exc:
            self.assertEqual(str(exc), "'extcode' <class ExternalCodeComp>: Error calling compute(), "
                             "Timed out after 1.0 sec.")
        else:
            self.fail('Expected AnalysisError')

    def test_error_code_raise(self):
        self.extcode.options['command'] = [
            sys.executable, 'extcode_example.py', 'extcode.out', '--delay', '-3'
        ]
        self.extcode.options['timeout'] = 1.0

        self.extcode.options['external_input_files'] = ['extcode_example.py']

        self.prob.setup(check=True)
        try:
            self.prob.run_model()
        except RuntimeError as exc:
            self.assertTrue('Traceback' in str(exc),
                            "no traceback found in '%s'" % str(exc))
            self.assertEqual(self.extcode.return_code, 1)
        else:
            self.fail('Expected RuntimeError')

    def test_error_code_soft(self):
        self.extcode.options['command'] = [
            sys.executable, 'extcode_example.py', 'extcode.out', '--delay', '-3'
        ]
        self.extcode.options['timeout'] = 1.0
        self.extcode.options['fail_hard'] = False

        self.extcode.options['external_input_files'] = ['extcode_example.py']

        self.prob.setup(check=True)
        try:
            self.prob.run_model()
        except om.AnalysisError as err:
            self.assertTrue("delay must be >= 0" in str(err),
                            "expected 'delay must be >= 0' to be in '%s'" % str(err))
            self.assertTrue('Traceback' in str(err),
                            "no traceback found in '%s'" % str(err))
        else:
            self.fail("AnalysisError expected")

    def test_allowed_return_code(self):
        self.extcode.options['allowed_return_codes'] = set(range(5))
        self.extcode.options['command'] = [
            sys.executable, 'extcode_example.py', 'extcode.out', '--return_code', '4'
        ]

        self.extcode.options['external_input_files'] = ['extcode_example.py']

        self.prob.setup(check=True)
        self.prob.run_model()

    def test_disallowed_return_code(self):
        self.extcode.options['allowed_return_codes'] = list(range(5))
        self.extcode.options['command'] = [
            sys.executable, 'extcode_example.py', 'extcode.out', '--return_code', '7'
        ]

        self.extcode.options['external_input_files'] = ['extcode_example.py']

        self.prob.setup(check=True)
        try:
            self.prob.run_model()
        except RuntimeError as err:
            self.assertTrue("return_code = 7" in str(err),
                            "expected 'return_code = 7' to be in '%s'" % str(err))
        else:
            self.fail("RuntimeError expected")

    def test_badcmd(self):
        # Set command to nonexistant path.
        self.extcode.options['command'] = ['no-such-command']

        self.prob.setup()
        try:
            self.prob.run_model()
        except ValueError as exc:
            self.assertEqual(str(exc), "'extcode' <class ExternalCodeComp>: Error calling compute(), "
                             "The command to be executed, 'no-such-command', cannot be found")
            self.assertEqual(self.extcode.return_code, -999999)
        else:
            self.fail('Expected ValueError')

    def test_nullcmd(self):
        self.extcode.stdout = 'nullcmd.out'
        self.extcode.stderr = STDOUT

        self.prob.setup()
        try:
            self.prob.run_model()
        except ValueError as exc:
            self.assertEqual(str(exc), "'extcode' <class ExternalCodeComp>: Error calling compute(), "
                             "Empty command list")
        else:
            self.fail('Expected ValueError')
        finally:
            if os.path.exists(self.extcode.stdout):
                os.remove(self.extcode.stdout)

    def test_env_vars(self):
        self.extcode.options['env_vars'] = {'TEST_ENV_VAR': 'SOME_ENV_VAR_VALUE'}
        self.extcode.options['command'] = [
            sys.executable, 'extcode_example.py', 'extcode.out', '--write_test_env_var'
        ]

        self.prob.setup(check=True)
        self.prob.run_model()

        # Check to see if output file contains the env var value
        with open(os.path.join(self.tempdir, 'extcode.out'), 'r') as out:
            file_contents = out.read()
        self.assertTrue('SOME_ENV_VAR_VALUE' in file_contents,
                        "'SOME_ENV_VAR_VALUE' missing from '%s'" % file_contents)


class TestExternalCodeCompArgs(unittest.TestCase):

    def test_kwargs(self):
        # check kwargs are passed to options
        extcode = om.ExternalCodeComp(poll_delay=999)

        self.assertTrue(extcode.options['poll_delay'] == 999)

        # check subclass kwargs are also passed to options
        class MyComp(om.ExternalCodeComp):
            def initialize(self):
                self.options.declare('my_arg', 'foo', desc='subclass option')

        my_comp = MyComp(poll_delay=999, my_arg='bar')

        self.assertTrue(my_comp.options['poll_delay'] == 999)
        self.assertTrue(my_comp.options['my_arg'] == 'bar')

        # check that options are those declared in both classes
        extcode_opts = set(extcode.options._dict.keys())
        my_comp_opts = set(my_comp.options._dict.keys())

        self.assertEqual(my_comp_opts.difference(extcode_opts), set(('my_arg',)))


class ParaboloidExternalCodeComp(om.ExternalCodeComp):
    def setup(self):
        self.add_input('x', val=0.0)
        self.add_input('y', val=0.0)

        self.add_output('f_xy', val=0.0)

        self.input_file = 'paraboloid_input.dat'
        self.output_file = 'paraboloid_output.dat'

        # providing these is optional; the component will verify that any input
        # files exist before execution and that the output files exist after.
        self.options['external_input_files'] = [self.input_file]
        self.options['external_output_files'] = [self.output_file]

        # If you want to write your command as a list, the code below will also work.
        # self.options['command'] = [
        #     sys.executable, 'extcode_paraboloid.py', self.input_file, self.output_file
        # ]

        self.options['command'] = ('python extcode_paraboloid.py {} {}').format(self.input_file, self.output_file)

    def compute(self, inputs, outputs):
        x = inputs['x'].item()
        y = inputs['y'].item()

        # generate the input file for the paraboloid external code
        with open(self.input_file, 'w') as input_file:
            input_file.write('%.16f\n%.16f\n' % (x, y))

        # the parent compute function actually runs the external code
        super().compute(inputs, outputs)

        # parse the output file from the external code and set the value of f_xy
        with open(self.output_file, 'r') as output_file:
            f_xy = float(output_file.read())

        outputs['f_xy'] = f_xy


class ParaboloidExternalCodeCompFD(om.ExternalCodeComp):
    def setup(self):
        self.add_input('x', val=0.0)
        self.add_input('y', val=0.0)

        self.add_output('f_xy', val=0.0)

        self.input_file = 'paraboloid_input.dat'
        self.output_file = 'paraboloid_output.dat'

        # providing these is optional; the component will verify that any input
        # files exist before execution and that the output files exist after.
        self.options['external_input_files'] = [self.input_file]
        self.options['external_output_files'] = [self.output_file]

        self.options['command'] = [
            sys.executable, 'extcode_paraboloid.py', self.input_file, self.output_file
        ]

    def setup_partials(self):
        # this external code does not provide derivatives, use finite difference
        self.declare_partials(of='*', wrt='*', method='fd')

    def compute(self, inputs, outputs):
        x = inputs['x'].item()
        y = inputs['y'].item()

        # generate the input file for the paraboloid external code
        with open(self.input_file, 'w') as input_file:
            input_file.write('%.16f\n%.16f\n' % (x, y))

        # the parent compute function actually runs the external code
        super().compute(inputs, outputs)

        # parse the output file from the external code and set the value of f_xy
        with open(self.output_file, 'r') as output_file:
            f_xy = float(output_file.read())

        outputs['f_xy'] = f_xy


class ParaboloidExternalCodeCompDerivs(om.ExternalCodeComp):
    def setup(self):
        self.add_input('x', val=0.0)
        self.add_input('y', val=0.0)

        self.add_output('f_xy', val=0.0)

        self.input_file = 'paraboloid_input.dat'
        self.output_file = 'paraboloid_output.dat'
        self.derivs_file = 'paraboloid_derivs.dat'

        # providing these is optional; the component will verify that any input
        # files exist before execution and that the output files exist after.
        self.options['external_input_files'] = [self.input_file]
        self.options['external_output_files'] = [self.output_file, self.derivs_file]

        self.options['command'] = [
            sys.executable, 'extcode_paraboloid_derivs.py',
            self.input_file, self.output_file, self.derivs_file
        ]

    def setup_partials(self):
        # this external code does provide derivatives
        self.declare_partials(of='*', wrt='*')

    def compute(self, inputs, outputs):
        x = inputs['x'].item()
        y = inputs['y'].item()

        # generate the input file for the paraboloid external code
        with open(self.input_file, 'w') as input_file:
            input_file.write('%.16f\n%.16f\n' % (x, y))

        # the parent compute function actually runs the external code
        super().compute(inputs, outputs)

        # parse the output file from the external code and set the value of f_xy
        with open(self.output_file, 'r') as output_file:
            f_xy = float(output_file.read())

        outputs['f_xy'] = f_xy

    def compute_partials(self, inputs, partials):
        outputs = {}

        # the parent compute function actually runs the external code
        super().compute(inputs, outputs)

        # parse the derivs file from the external code and set partials
        with open(self.derivs_file, 'r') as derivs_file:
            partials['f_xy', 'x'] = float(derivs_file.readline())
            partials['f_xy', 'y'] = float(derivs_file.readline())


class TestExternalCodeCompFeature(unittest.TestCase):

    def setUp(self):
        import os
        import shutil
        import tempfile

        # get the directory where the needed support files are located
        import openmdao.components.tests.test_external_code_comp as extcode_test
        DIRECTORY = os.path.dirname((os.path.abspath(extcode_test.__file__)))

        # change to temp dir
        self.startdir = os.getcwd()
        self.tempdir = tempfile.mkdtemp(prefix='test_extcode-')
        os.chdir(self.tempdir)

        # copy required files to temp dir
        files = ['extcode_paraboloid.py', 'extcode_paraboloid_derivs.py']
        for filename in files:
            shutil.copy(os.path.join(DIRECTORY, filename),
                        os.path.join(self.tempdir, filename))

    def tearDown(self):
        # destroy the evidence
        os.chdir(self.startdir)
        try:
            shutil.rmtree(self.tempdir)
        except OSError:
            pass

    def test_main(self):

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('p', ParaboloidExternalCodeComp(), promotes_inputs=['x', 'y'])

        # run the ExternalCodeComp Component
        prob.setup()

        # Set input values
        prob.set_val('p.x', 3.0)
        prob.set_val('p.y', -4.0)

        prob.run_model()

        # print the output
        self.assertEqual(prob.get_val('p.f_xy'), -15.0)

    def test_optimize_fd(self):

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('p', ParaboloidExternalCodeCompFD())

        # find optimal solution with SciPy optimize
        # solution (minimum): x = 6.6667; y = -7.3333
        prob.driver = om.ScipyOptimizeDriver()
        prob.driver.options['optimizer'] = 'SLSQP'

        prob.model.add_design_var('p.x', lower=-50, upper=50)
        prob.model.add_design_var('p.y', lower=-50, upper=50)

        prob.model.add_objective('p.f_xy')

        prob.driver.options['tol'] = 1e-9
        prob.driver.options['disp'] = True

        prob.setup()

        # Set input values
        prob.set_val('p.x', 3.0)
        prob.set_val('p.y', -4.0)

        prob.run_driver()

        assert_near_equal(prob.get_val('p.x'), 6.66666667, 1e-6)
        assert_near_equal(prob.get_val('p.y'), -7.3333333, 1e-6)

    def test_optimize_derivs(self):

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('p', ParaboloidExternalCodeCompDerivs())

        # find optimal solution with SciPy optimize
        # solution (minimum): x = 6.6667; y = -7.3333
        prob.driver = om.ScipyOptimizeDriver()
        prob.driver.options['optimizer'] = 'SLSQP'

        prob.model.add_design_var('p.x', lower=-50, upper=50)
        prob.model.add_design_var('p.y', lower=-50, upper=50)

        prob.model.add_objective('p.f_xy')

        prob.driver.options['tol'] = 1e-9
        prob.driver.options['disp'] = True

        prob.setup()

        # Set input values
        prob.set_val('p.x', 3.0)
        prob.set_val('p.y', -4.0)

        prob.run_driver()

        assert_near_equal(prob.get_val('p.x'), 6.66666667, 1e-6)
        assert_near_equal(prob.get_val('p.y'), -7.3333333, 1e-6)


class TestExternalCodeImplicitCompFeature(unittest.TestCase):

    def setUp(self):

        # get the directory where the needed support files are located
        import openmdao.components.tests.test_external_code_comp as extcode_test
        DIRECTORY = os.path.dirname((os.path.abspath(extcode_test.__file__)))

        # change to temp dir
        self.startdir = os.getcwd()
        self.tempdir = tempfile.mkdtemp(prefix='test_extcode-')
        os.chdir(self.tempdir)

        # copy required files to temp dir
        files = ['extcode_resistor.py', 'extcode_node.py', 'extcode_mach.py']
        for filename in files:
            shutil.copy(os.path.join(DIRECTORY, filename),
                        os.path.join(self.tempdir, filename))

    def tearDown(self):
        # destroy the evidence
        os.chdir(self.startdir)
        try:
            shutil.rmtree(self.tempdir)
        except OSError:
            pass

    def test_simple_external_code_implicit_comp(self):

        class MachExternalCodeComp(om.ExternalCodeImplicitComp):

            def initialize(self):
                self.options.declare('super_sonic', types=bool)

            def setup(self):
                self.add_input('area_ratio', val=1.0, units=None)
                self.add_output('mach', val=1., units=None)

                self.input_file = 'mach_input.dat'
                self.output_file = 'mach_output.dat'

                # providing these are optional; the component will verify that any input
                # files exist before execution and that the output files exist after.
                self.options['external_input_files'] = [self.input_file]
                self.options['external_output_files'] = [self.output_file]


                self.options['command_apply'] = [
                    sys.executable, 'extcode_mach.py', self.input_file, self.output_file,
                ]
                self.options['command_solve'] = [
                    sys.executable, 'extcode_mach.py', self.input_file, self.output_file,
                ]

                # If you want to write your own string command, the code below will also work.
                # self.options['command_apply'] = ('python extcode_mach.py {} {}').format(self.input_file, self.output_file)

            def setup_partials(self):
                self.declare_partials(of='mach', wrt='area_ratio', method='fd')

            def apply_nonlinear(self, inputs, outputs, residuals):
                with open(self.input_file, 'w') as input_file:
                    input_file.write('residuals\n')
                    input_file.write('{}\n'.format(inputs['area_ratio'][0]))
                    input_file.write('{}\n'.format(outputs['mach'][0]))

                # the parent apply_nonlinear function actually runs the external code
                super().apply_nonlinear(inputs, outputs, residuals)

                # parse the output file from the external code and set the value of mach
                with open(self.output_file, 'r') as output_file:
                    mach = float(output_file.read())
                residuals['mach'] = mach

            def solve_nonlinear(self, inputs, outputs):
                with open(self.input_file, 'w') as input_file:
                    input_file.write('outputs\n')
                    input_file.write('{}\n'.format(inputs['area_ratio'][0]))
                    input_file.write('{}\n'.format(self.options['super_sonic']))
                # the parent apply_nonlinear function actually runs the external code
                super().solve_nonlinear(inputs, outputs)

                # parse the output file from the external code and set the value of mach
                with open(self.output_file, 'r') as output_file:
                    mach = float(output_file.read())
                outputs['mach'] = mach

        group = om.Group()
        mach_comp = group.add_subsystem('comp', MachExternalCodeComp(), promotes=['*'])
        prob = om.Problem(model=group)
        group.nonlinear_solver = om.NewtonSolver()
        group.nonlinear_solver.options['solve_subsystems'] = True
        group.nonlinear_solver.options['iprint'] = 0
        group.nonlinear_solver.options['maxiter'] = 20
        group.linear_solver = om.DirectSolver()

        prob.setup()

        area_ratio = 1.3
        super_sonic = False
        prob.set_val('area_ratio', area_ratio)
        mach_comp.options['super_sonic'] = super_sonic
        prob.run_model()
        assert_near_equal(prob.get_val('mach'), mach_solve(area_ratio, super_sonic=super_sonic), 1e-8)

        area_ratio = 1.3
        super_sonic = True
        prob.set_val('area_ratio', area_ratio)
        mach_comp.options['super_sonic'] = super_sonic
        prob.run_model()
        assert_near_equal(prob.get_val('mach'), mach_solve(area_ratio, super_sonic=super_sonic), 1e-8)

if __name__ == "__main__":
    unittest.main()




import unittest
import numpy as np

import openmdao.api as om

from openmdao.utils.mpi import MPI
from openmdao.utils.assert_utils import assert_near_equal, assert_check_totals

if MPI:
    from openmdao.api import PETScVector
    vector_class = PETScVector
    try:
        from pyoptsparse import Optimization as pyoptsparse_opt
    except ImportError:
        pyoptsparse_opt = None

else:
    PETScVector = None
    pyoptsparse_opt = None


class Mygroup(om.Group):

    def setup(self):
        self.add_subsystem('indep_var_comp', om.IndepVarComp('x'), promotes=['*'])
        self.add_subsystem('Cy', om.ExecComp('y=2*x'), promotes=['*'])
        self.add_subsystem('Cc', om.ExecComp('c=x+2'), promotes=['*'])

        self.add_design_var('x')
        self.add_constraint('c', lower=-3.)


@unittest.skipUnless(MPI and PETScVector and pyoptsparse_opt,
                     "MPI, PETSc and pyoptsparse are required.")
class RemoteVOITestCase(unittest.TestCase):

    N_PROCS = 2

    def test_remote_voi(self):
        prob = om.Problem()

        par = prob.model.add_subsystem('par', om.ParallelGroup())

        prob.model.par.add_subsystem('G1', Mygroup())
        prob.model.par.add_subsystem('G2', Mygroup())

        prob.model.add_subsystem('Obj', om.ExecComp('obj=y1+y2'))

        prob.model.connect('par.G1.y', 'Obj.y1')
        prob.model.connect('par.G2.y', 'Obj.y2')

        prob.model.add_objective('Obj.obj')

        prob.driver = om.pyOptSparseDriver()
        prob.driver.options['optimizer'] = 'SLSQP'
        prob.setup()

        self.assertEqual('par.G1', par.G1.pathname)
        self.assertEqual('par.G2', par.G2.pathname)

        prob.run_driver()

        J = prob.compute_totals(of=['Obj.obj', 'par.G1.c', 'par.G2.c'],
                                wrt=['par.G1.x', 'par.G2.x'])

        assert_near_equal(J['Obj.obj', 'par.G1.x'], np.array([[2.0]]), 1e-6)
        assert_near_equal(J['Obj.obj', 'par.G2.x'], np.array([[2.0]]), 1e-6)
        assert_near_equal(J['par.G1.c', 'par.G1.x'], np.array([[1.0]]), 1e-6)
        assert_near_equal(J['par.G1.c', 'par.G2.x'], np.array([[0.0]]), 1e-6)
        assert_near_equal(J['par.G2.c', 'par.G1.x'], np.array([[0.0]]), 1e-6)
        assert_near_equal(J['par.G2.c', 'par.G2.x'], np.array([[1.0]]), 1e-6)

    def test_check_totals_matfree_w_remote(self):
        # check_totals was hanging in some cases when all variables of interest didn't exist on
        # all procs. (Issue #2884)
        class DummyComp(om.ExplicitComponent):
            def initialize(self):
                self.options.declare('a',default=0.)
                self.options.declare('b',default=0.)

            def setup(self):
                self.add_input('x')
                self.add_output('y', 0.)

            def compute(self, inputs, outputs):
                outputs['y'] = self.options['a']*inputs['x'] + self.options['b']

            def compute_jacvec_product(self, inputs, d_inputs, d_outputs, mode):
                if mode=='rev':
                    if 'y' in d_outputs:
                        if 'x' in d_inputs:
                            d_inputs['x'] += self.options['a'] * d_outputs['y']

        class DummyGroup(om.ParallelGroup):
            def setup(self):
                self.add_subsystem('comp1',DummyComp(a=1,b=2.))
                self.add_subsystem('comp2',DummyComp(a=3.,b=4.))

        class Top(om.Group):
            def setup(self):
                self.add_subsystem('dvs',om.IndepVarComp(), promotes=['*'])
                self.dvs.add_output('x',1.)
                self.add_subsystem('parallel_group',DummyGroup())
                self.connect('x','parallel_group.comp1.x')
                self.connect('x','parallel_group.comp2.x')

        prob = om.Problem(model=Top())
        prob.setup(mode='rev')
        prob.run_model()

        data = prob.check_totals(of='parallel_group.comp1.y',wrt='x', out_stream=None)
        assert_check_totals(data, atol=1e-6, rtol=1e-6)


if __name__ == "__main__":
    from openmdao.utils.mpi import mpirun_tests
    mpirun_tests()



"""Test for a feature doc showing how to use cache_linear_solution"""
import unittest

import numpy as np
from scipy.sparse.linalg import gmres

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal


class CacheLinearTestCase(unittest.TestCase):

    def test_feature_cache_linear(self):

        class QuadraticComp(om.ImplicitComponent):
            """
            A Simple Implicit Component representing a Quadratic Equation.

            R(a, b, c, x) = ax^2 + bx + c

            Solution via Quadratic Formula:
            x = (-b + sqrt(b^2 - 4ac)) / 2a
            """

            def setup(self):
                self.add_input('a', val=1.)
                self.add_input('b', val=1.)
                self.add_input('c', val=1.)
                self.add_output('states', val=[0,0])

            def setup_partials(self):
                self.declare_partials(of='*', wrt='*')

            def apply_nonlinear(self, inputs, outputs, residuals):
                a = inputs['a'].item()
                b = inputs['b'].item()
                c = inputs['c'].item()
                x = outputs['states'][0]
                y = outputs['states'][1]

                residuals['states'][0] = a * x ** 2 + b * x + c
                residuals['states'][1] = a * y + b

            def solve_nonlinear(self, inputs, outputs):
                a = inputs['a'].item()
                b = inputs['b'].item()
                c = inputs['c'].item()
                outputs['states'][0] = (-b + (b ** 2 - 4 * a * c) ** 0.5) / (2 * a)
                outputs['states'][1] = -b/a

            def linearize(self, inputs, outputs, partials):
                a = inputs['a'][0]
                b = inputs['b'][0]
                c = inputs['c'][0]
                x = outputs['states'][0]
                y = outputs['states'][1]

                partials['states', 'a'] = [[x**2],[y]]
                partials['states', 'b'] = [[x],[1]]
                partials['states', 'c'] = [[1.0],[0]]
                partials['states', 'states'] = [[2*a*x+b, 0],[0, a]]

                self.state_jac = np.array([[2*a*x+b, 0],[0, a]])

            def solve_linear(self, d_outputs, d_residuals, mode):

                if mode == 'fwd':
                    d_outputs['states'] = gmres(self.state_jac, d_residuals['states'], x0=d_outputs['states'], atol=0)[0]

                elif mode == 'rev':
                    d_residuals['states'] = gmres(self.state_jac, d_outputs['states'], x0=d_residuals['states'], atol=0)[0]

        p = om.Problem()
        p.driver = om.ScipyOptimizeDriver()
        p.driver.options['optimizer'] = 'SLSQP'

        p.model.set_input_defaults('b', val=4.)
        p.model.add_subsystem('quad', QuadraticComp(), promotes_inputs=['a', 'b', 'c'], promotes_outputs=['states'])
        p.model.add_subsystem('obj', om.ExecComp('y = (x[1]-x[0])**2', x=np.ones(2)))
        p.model.connect('states', 'obj.x')

        p.model.add_design_var('a', cache_linear_solution=True, upper=4.)
        p.model.add_constraint('states', upper=10)
        p.model.add_objective('obj.y')

        p.setup(mode='fwd')
        p.run_driver()

        print(p['a'], p['b'], p['c'])
        print(p['states'])
        assert_near_equal(p['obj.y'], 0.25000053, 1e-3)


if __name__ == "__main__":
    unittest.main()




"""
Define the BroydenSolver class.

Based on implementation in Scipy via OpenMDAO 0.8x with improvements based on NPSS solver.
"""
import numpy as np

from openmdao.recorders.recording_iteration_stack import Recording
from openmdao.solvers.linesearch.backtracking import BoundsEnforceLS
from openmdao.solvers.solver import NonlinearSolver
from openmdao.utils.class_util import overrides_method
from openmdao.utils.om_warnings import issue_warning, SetupWarning
from openmdao.utils.mpi import MPI


CITATION = """@ARTICLE{
              Broyden1965ACo,
              AUTHOR = "C. Broyden",
              TITLE = "A Class of Methods for Solving Nonlinear Simultaneous Equations",
              JOURNAL = "Mathematics of Computation",
              VOLUME = "19",
              YEAR = "1965",
              PAGES = "577--593",
              REFERRED = "[Coleman1996SaE]."
              }"""


class BroydenSolver(NonlinearSolver):
    """
    Broyden solver.

    Parameters
    ----------
    **kwargs : dict
        Options dictionary.

    Attributes
    ----------
    delta_fxm : ndarray
        Most recent change in residual vector.
    delta_xm : ndarray
        Most recent change in state vector.
    fxm : ndarray
        Most recent residual.
    Gm : ndarray
        Most recent Jacobian matrix.
    linear_solver : LinearSolver
        Linear solver to use for calculating inverse Jacobian.
    _linesearch : NonlinearSolver
        Line search algorithm. Default is None for no line search.
    size : int
        Total length of the states being solved.
    xm : ndarray
        Most recent state.
    _idx : dict
        Cache of vector indices for each state name.
    _computed_jacobians : int
        Number of computed jacobians.
    _converge_failures : int
        Number of consecutive iterations that failed to converge to the tol definied in options.
    _full_inverse : bool
        When True, Broyden considers the whole vector rather than a list of states.
    _recompute_jacobian : bool
        Flag that becomes True when Broyden detects it needs to recompute the inverse Jacobian.
    """

    SOLVER = 'NL: BROYDEN'

    def __init__(self, **kwargs):
        """
        Initialize all attributes.
        """
        super().__init__(**kwargs)

        # Slot for linear solver
        self.linear_solver = None

        # Slot for linesearch
        self.supports['linesearch'] = True
        self._linesearch = BoundsEnforceLS()

        self.cite = CITATION

        self.size = 0
        self._idx = {}
        self._recompute_jacobian = True
        self.Gm = None
        self.xm = None
        self.fxm = None
        self.delta_xm = None
        self.delta_fxm = None
        self._converge_failures = 0
        self._computed_jacobians = 0

        # This gets set to True if the user doesn't declare any states.
        self._full_inverse = False

    def _declare_options(self):
        """
        Declare options before kwargs are processed in the init method.
        """
        super()._declare_options()

        self.options.declare('alpha', default=0.4,
                             desc="Value to scale the starting Jacobian, which is Identity. This "
                                  "option does nothing if you compute the initial Jacobian "
                                  "instead.")
        self.options.declare('compute_jacobian', types=bool, default=True,
                             desc="When True, compute an initial Jacobian, otherwise start "
                                  "with Identity scaled by alpha. Further Jacobians may also be "
                                  "computed depending on the other options.")
        self.options.declare('converge_limit', default=1.0,
                             desc="Ratio of current residual to previous residual above which the "
                                  "convergence is considered a failure. The Jacobian will be "
                                  "regenerated once this condition has been reached a number of "
                                  "consecutive times as specified in max_converge_failures.")
        self.options.declare('cs_reconverge', types=bool, default=True,
                             desc='When True, when this driver solves under a complex step, nudge '
                             'the Solution vector by a small amount so that it reconverges.')
        self.options.declare('diverge_limit', default=2.0,
                             desc="Ratio of current residual to previous residual above which the "
                                  "Jacobian will be immediately regenerated.")
        self.options.declare('max_converge_failures', default=3,
                             desc="The number of convergence failures before regenerating the "
                                  "Jacobian.")
        self.options.declare('max_jacobians', default=10,
                             desc="Maximum number of jacobians to compute.")
        self.options.declare('state_vars', [], desc="List of the state-variable/residuals that "
                                                    "are to be solved here.")
        self.options.declare('update_broyden', default=True,
                             desc="Flag controls whether to perform Broyden update to the "
                                  "Jacobian. There are some applications where it may be useful "
                                  "to turn this off.")
        self.options.declare('reraise_child_analysiserror', types=bool, default=False,
                             desc='When the option is true, a solver will reraise any '
                             'AnalysisError that arises during subsolve; when false, it will '
                             'continue solving.')

        self.supports['gradients'] = True
        self.supports['implicit_components'] = True

    def _setup_solvers(self, system, depth):
        """
        Assign system instance, set depth, and optionally perform setup.

        Parameters
        ----------
        system : <System>
            Pointer to the owning system.
        depth : int
            Depth of the current system (already incremented).
        """
        super()._setup_solvers(system, depth)
        self._recompute_jacobian = True
        self._computed_jacobians = 0

        self._disallow_discrete_outputs()

        if self.linear_solver is not None:
            self.linear_solver._setup_solvers(system, self._depth + 1)
        else:
            self.linear_solver = system.linear_solver

        if self.linesearch is not None:
            self.linesearch._setup_solvers(system, self._depth + 1)
            self.linesearch._do_subsolve = True

        # this check is incorrect (for broyden) and needs to be done differently.
        # self._disallow_distrib_solve()

        states = self.options['state_vars']
        prom2abs = system._var_allprocs_prom2abs_list['output']

        # Check names of states.
        bad_names = [name for name in states if name not in prom2abs]
        if len(bad_names) > 0:
            msg = "{}: The following variable names were not found: {}"
            raise ValueError(msg.format(self.msginfo, ', '.join(bad_names)))

        # Size linear system
        if len(states) > 0:
            # User has specified states, so we must size them.
            n = 0
            meta = system._var_allprocs_abs2meta['output']

            for i, name in enumerate(states):
                size = meta[prom2abs[name][0]]['global_size']
                self._idx[name] = (n, n + size)
                n += size
        else:
            # Full system size.
            self._full_inverse = True
            n = np.sum(system._owned_sizes)

        self.size = n
        self.Gm = np.empty((n, n))
        self.xm = np.empty((n, ))
        self.fxm = np.empty((n, ))
        self.delta_xm = None
        self.delta_fxm = None

        if self._full_inverse:

            # Can only use DirectSolver here.
            from openmdao.solvers.linear.direct import DirectSolver
            if not isinstance(self.linear_solver, DirectSolver):
                msg = "{}: Linear solver must be DirectSolver when solving the full model."
                raise ValueError(msg.format(self.msginfo, ', '.join(bad_names)))

            return

        # Always look for states that aren't being solved so we can warn the user.
        def sys_recurse(system, all_states):
            subs = system._subsystems_myproc
            if len(subs) == 0:

                # Skip implicit components that appear to solve themselves.
                from openmdao.core.implicitcomponent import ImplicitComponent
                if overrides_method('solve_nonlinear', system, ImplicitComponent):
                    return

                all_states.extend(system._list_states())

            else:
                for subsys in subs:
                    sub_nl = subsys.nonlinear_solver
                    if sub_nl and sub_nl.supports['implicit_components']:
                        continue
                    sys_recurse(subsys, all_states)

        all_states = []
        sys_recurse(system, all_states)
        all_states = [system._var_abs2prom['output'][name] for name in all_states]

        missing = set(all_states).difference(states)
        if len(missing) > 0:
            msg = "The following states are not covered by a solver, and may have been " + \
                  "omitted from the BroydenSolver 'state_vars': "
            msg += ', '.join(sorted(missing))
            issue_warning(msg, category=SetupWarning)

    def _assembled_jac_solver_iter(self):
        """
        Return a generator of linear solvers using assembled jacs.
        """
        if self.linear_solver is not None:
            for s in self.linear_solver._assembled_jac_solver_iter():
                yield s

    def _set_solver_print(self, level=2, type_='all'):
        """
        Control printing for solvers and subsolvers in the model.

        Parameters
        ----------
        level : int
            iprint level. Set to 2 to print residuals each iteration; set to 1
            to print just the iteration totals; set to 0 to disable all printing
            except for failures, and set to -1 to disable all printing including failures.
        type_ : str
            Type of solver to set: 'LN' for linear, 'NL' for nonlinear, or 'all' for all.
        """
        super()._set_solver_print(level=level, type_=type_)

        if self.linear_solver is not None and type_ != 'NL':
            self.linear_solver._set_solver_print(level=level, type_=type_)

        if self.linesearch is not None:
            self.linesearch._set_solver_print(level=level, type_=type_)

    def _linearize(self):
        """
        Perform any required linearization operations such as matrix factorization.
        """
        if self.linear_solver is not None:
            self.linear_solver._linearize()

        if self.linesearch is not None:
            self.linesearch._linearize()

    def _iter_initialize(self):
        """
        Perform any necessary pre-processing operations.

        Returns
        -------
        float
            Initial relative error in the user-specified residuals.
        float
            Initial absolute error in the user-specified residuals.
        """
        system = self._system()
        if self.options['debug_print']:
            self._err_cache['inputs'] = system._inputs._copy_views()
            self._err_cache['outputs'] = system._outputs._copy_views()

        # Convert local storage if we are under complex step.
        if system.under_complex_step:
            self.Gm = self.Gm.astype(complex)
            self.xm = self.xm.astype(complex)
            self.fxm = self.fxm.astype(complex)
        elif np.iscomplexobj(self.xm):
            self.Gm = self.Gm.real
            self.xm = self.xm.real
            self.fxm = self.fxm.real

        self._converge_failures = 0
        self._computed_jacobians = 0

        # Execute guess_nonlinear if specified and
        # we have not restarted from a saved point
        if not self._restarted and system._has_guess:
            system._guess_nonlinear()

        # When under a complex step from higher in the hierarchy, sometimes the step is too small
        # to trigger reconvergence, so nudge the outputs slightly so that we always get at least
        # one iteration of Broyden.
        if system.under_complex_step and self.options['cs_reconverge']:
            system._outputs += np.linalg.norm(system._outputs.asarray()) * 1e-10

        # Start with initial states.
        self.xm = self.get_vector(system._outputs)

        with Recording('Broyden', 0, self) as rec:
            self._solver_info.append_solver()

            # should call the subsystems solve before computing the first residual
            self._gs_iter()

            self._solver_info.pop()

            self._run_apply()
            norm = self._iter_get_norm()

            rec.abs = norm
            norm0 = norm if norm != 0.0 else 1.0
            rec.rel = norm / norm0

        return norm0, norm

    def _iter_get_norm(self):
        """
        Return the norm of only the residuals requested in options.

        Returns
        -------
        float
            Norm of the residuals.
        """
        # Need to cache the initial residuals, which is done in this function.
        self.fxm = fxm = self.get_vector(self._system()._residuals)
        if not self._full_inverse:
            # Use full model residual for driving the main loop convergence.
            fxm = self._system()._residuals.asarray()

        return self.compute_norm(fxm)

    def compute_norm(self, vec):
        """
        Compute norm of the vector.

        Under MPI, compute the norm on rank 0, and broadcast it to all other ranks.

        Parameters
        ----------
        vec : ndarray
            Array of real or complex values. For MPI on rank 0, should be full dimension of the
            openmdao vector with duplicate indices removed.

        Returns
        -------
        float
            Norm of vec, computed on rank 0 and broadcast to all other ranks.
        """
        return np.linalg.norm(vec)

    def _single_iteration(self):
        """
        Perform the operations in the iteration loop.
        """
        system = self._system()
        Gm = self._update_inverse_jacobian()
        fxm = self.fxm

        delta_xm = -Gm.dot(fxm)

        if self.linesearch:
            self._solver_info.append_subsolver()

            self.set_states(self.xm)
            self.set_linear_vector(delta_xm)
            self.linesearch.solve()
            xm = self.get_vector(system._outputs)

            self._solver_info.pop()

        else:
            # Update the new states in the model.
            xm = self.xm + delta_xm
            self.set_states(xm)

        # Run the model.
        self._solver_info.append_solver()
        self._gs_iter()
        self._solver_info.pop()

        self._run_apply()

        fxm1 = fxm.copy()
        self.fxm = fxm = self.get_vector(system._residuals)
        delta_fxm = fxm - fxm1

        # States may have been further converged hierarchically.
        xm = self.get_vector(system._outputs)
        delta_xm = xm - self.xm

        # Determine whether to update Jacobian.
        self._recompute_jacobian = False
        opt = self.options
        if self._computed_jacobians <= opt['max_jacobians']:

            converge_ratio = self.compute_norm(fxm) / self.compute_norm(fxm1)

            if converge_ratio > opt['diverge_limit']:
                self._recompute_jacobian = True
            elif converge_ratio > opt['converge_limit']:
                self._converge_failures += 1

                if self._converge_failures >= opt['max_converge_failures']:
                    self._recompute_jacobian = True
            else:
                self._converge_failures = 0

        # Cache for next iteration.
        self.delta_xm = delta_xm
        self.delta_fxm = delta_fxm
        self.fxm = fxm
        self.xm = xm
        self.Gm = Gm

    def _update_inverse_jacobian(self):
        """
        Update the inverse Jacobian for a new Broyden iteration.

        Returns
        -------
        ndarray
            Updated inverse Jacobian.
        """
        Gm = self.Gm

        # Apply the Broyden Update approximation to the previous value of the inverse jacobian.
        if self.options['update_broyden'] and not self._recompute_jacobian:
            dfxm = self.delta_fxm
            fact = np.linalg.norm(dfxm)

            # Sometimes you can get stuck, particularly when enforcing bounds in a linesearch.
            # Make sure we don't update in this case because of divide by zero.
            if fact > self.options['atol']:
                Gm += np.outer((self.delta_xm - Gm.dot(dfxm)), dfxm * (1.0 / fact**2))

        # Solve for total derivatives of user-requested residuals wrt states.
        elif self.options['compute_jacobian']:
            if self._full_inverse:
                Gm = self._compute_full_inverse_jacobian()
            else:
                Gm = self._compute_inverse_jacobian()

            self._computed_jacobians += 1

        # Set inverse Jacobian to identity scaled by alpha.
        # This is the default starting point used by scipy and the general broyden algorithm.
        else:
            Gm = np.diag(np.full(self.size, -self.options['alpha'], dtype=Gm.dtype))

        return Gm

    def get_vector(self, vec):
        """
        Return a vector containing the values of vec at the states specified in options.

        This is the full incoming vec when no states are defined. When under MPI, the values are
        appopriately gathered without duplicates to rank 0.

        Parameters
        ----------
        vec : <Vector>
            Vector from which to extract state values.

        Returns
        -------
        ndarray
            Array containing values of vector at desired states.
        """
        if self._full_inverse:
            xm = vec.asarray(copy=True)
        else:
            states = self.options['state_vars']
            xm = self.xm.copy()
            for name in states:
                i, j = self._idx[name]
                xm[i:j] = vec[name]

        return xm

    def set_states(self, new_val):
        """
        Set new values for states specified in options.

        Parameters
        ----------
        new_val : ndarray
            New values for states.
        """
        outputs = self._system()._outputs

        if self._full_inverse:
            outputs.set_val(new_val)
        else:
            states = self.options['state_vars']
            for name in states:
                i, j = self._idx[name]
                outputs[name] = new_val[i:j]

    def set_linear_vector(self, dx):
        """
        Copy values from step into the linear vector for backtracking.

        Parameters
        ----------
        dx : ndarray
            Full step in the states for this iteration.
        """
        linear = self._system()._doutputs

        if self._full_inverse:
            linear.set_val(dx)
        else:
            linear.set_val(0.0)
            for name in self.options['state_vars']:
                i, j = self._idx[name]
                linear[name] = dx[i:j]

    def _compute_inverse_jacobian(self):
        """
        Compute inverse Jacobian for system by doing a linear solve for each state.

        Returns
        -------
        ndarray
            New inverse Jacobian.
        """
        # TODO: Consider promoting this capability out into OpenMDAO so other solvers can use the
        # same code.
        # TODO: Can do each state in parallel if procs are available.
        system = self._system()
        states = self.options['state_vars']
        d_res = system._dresiduals
        d_out = system._doutputs

        inv_jac = self.Gm
        d_res.set_val(0.0)

        # Disable local fd
        approx_status = system._owns_approx_jac
        system._owns_approx_jac = False

        try:
            # Linearize model.
            ln_solver = self.linear_solver
            do_sub_ln = ln_solver._linearize_children()
            my_asm_jac = ln_solver._assembled_jac
            system._linearize(my_asm_jac, sub_do_ln=do_sub_ln)
            if my_asm_jac is not None and system.linear_solver._assembled_jac is not my_asm_jac:
                my_asm_jac._update(system)
            self._linearize()

            for wrt_name in states:
                i_wrt, j_wrt = self._idx[wrt_name]
                if wrt_name in d_res:
                    d_wrt = d_res[wrt_name]

                for j in range(j_wrt - i_wrt):

                    # Increment each variable.
                    if wrt_name in d_res:
                        d_wrt[j] = 1.0

                    # Solve for total derivatives.
                    ln_solver.solve('fwd')

                    # Extract results.
                    for of_name in states:
                        i_of, j_of = self._idx[of_name]
                        inv_jac[i_of:j_of, i_wrt + j] = d_out[of_name]

                    if wrt_name in d_res:
                        d_wrt[j] = 0.0
        finally:
            # Enable local fd
            system._owns_approx_jac = approx_status

        return inv_jac

    def _compute_full_inverse_jacobian(self):
        """
        Compute inverse Jacobian for entire system vector.

        Only the DirectSolver is supported here.

        Returns
        -------
        ndarray
            New inverse Jacobian.
        """
        system = self._system()

        # Disable local fd
        approx_status = system._owns_approx_jac
        system._owns_approx_jac = False

        try:
            # Linearize model.
            ln_solver = self.linear_solver
            do_sub_ln = ln_solver._linearize_children()
            my_asm_jac = ln_solver._assembled_jac
            system._linearize(my_asm_jac, sub_do_ln=do_sub_ln)
            if my_asm_jac is not None and system.linear_solver._assembled_jac is not my_asm_jac:
                my_asm_jac._update(system)

            inv_jac = self.linear_solver._inverse()
        finally:
            # Enable local fd
            system._owns_approx_jac = approx_status

        return inv_jac

    def cleanup(self):
        """
        Clean up resources prior to exit.
        """
        super().cleanup()

        if self.linear_solver:
            self.linear_solver.cleanup()
        if self.linesearch:
            self.linesearch.cleanup()



"""
Unit tests for the spline distribution class.
"""
import unittest

import numpy as np
from numpy.testing import assert_array_almost_equal

import openmdao.api as om
from openmdao.utils.spline_distributions import cell_centered, sine_distribution, node_centered


class SplineDistributionsTestCase(unittest.TestCase):

    def test_cell_centered(self):

        calculated_midpoints = np.array([2.1, 4.3, 6.5, 8.7, 10.9])

        dist = cell_centered(5, start=1.0, end=12.0)

        assert_array_almost_equal(calculated_midpoints, dist)

    def test_sin_distribution(self):

        calculated = np.array([0.        , 0.03015369, 0.11697778, 0.25      , 0.41317591,
                               0.58682409, 0.75      , 0.88302222, 0.96984631, 1.        ])

        dist = sine_distribution(10)

        assert_array_almost_equal(calculated, dist)

        calculated = np.array([0.14644661, 0.21321178, 0.28869087, 0.37059048, 0.45642213,
                               0.54357787, 0.62940952, 0.71130913, 0.78678822, 0.85355339])


        dist = sine_distribution(10, phase=np.pi/2.0)

        assert_array_almost_equal(calculated, dist)

    def test_node_centered(self):

        calculated_midpoints = np.array([1., 1.3, 1.6, 1.9, 2.2, 2.5, 2.8, 3.1, 3.4, 3.7, 4.])

        assert_array_almost_equal(calculated_midpoints, node_centered(10, start=1.0, end=4.0))


class SplineDistributionFeatureTestCase(unittest.TestCase):

    def test_spline_distribution_example(self):

        x_cp = np.linspace(0., 1., 6)
        y_cp = np.array([5.0, 12.0, 14.0, 16.0, 21.0, 29.0])
        n = 20
        x = sine_distribution(n, start=0.0, end=1.0, phase=np.pi)

        prob = om.Problem()

        comp = om.SplineComp(method='akima', x_cp_val=x_cp, x_interp_val=x)
        prob.model.add_subsystem('akima1', comp)

        comp.add_spline(y_cp_name='ycp', y_interp_name='y_val', y_cp_val=y_cp)

        prob.setup(force_alloc_complex=True)
        prob.run_model()

        assert_array_almost_equal(prob['akima1.y_val'],
                                  np.array([[5.         , 5.32381994,  6.28062691 , 7.79410646 , 9.64169506, 11.35166363,
                                             12.26525921, 12.99152288, 13.77257256, 14.58710327, 15.41289673, 16.28341046,
                                             17.96032258, 20.14140712, 22.31181718, 24.40891577, 26.27368825, 27.74068235,
                                             28.67782484, 29.        ]])                                  )


if __name__ == '__main__':
    unittest.main()


"""
Table building classes.
"""

import sys
import os
import json
import textwrap
from collections import namedtuple
from itertools import zip_longest, chain
from html import escape
from dataclasses import dataclass
from numbers import Number, Integral
from openmdao.utils.notebook_utils import notebook, display, HTML, IFrame, colab
from openmdao.utils.om_warnings import issue_warning


_align2symbol = {
    'center': '^',
    'right': '>',
    'left': '<'
}

_default_align = {
    'int': 'right',
    'real': 'right',
    'bool': 'center',
    'other': 'left',
}

_big_table_height = 600
_row_height_px = 45  # use to size the IFrame in a notebook


def _num_cols(rows):
    ncols = None
    for row in rows:
        i = 0
        for _ in row:
            i += 1
        if ncols is None or ncols < i:
            ncols = i
    return 0 if ncols is None else ncols


class TableBuilder(object):
    """
    Base class for all table builders.

    Parameters
    ----------
    rows : iter of iters
        Data used to fill table cells.
    headers : iter of str, 'keys', or None
        If not None, header strings for all columns.  Size must match
        number of columns in each row of data in rows.  A value of 'keys' means that rows is
        either a dict or a list of dicts and the keys of that/those dict(s) should be used
        as headers.
    column_meta : iter of dict or None
        If not None, contains a dict for each table column and the dict contains values
        of metadata for that column.
    precision : int or str
        Precision applied to all columns of real numbers.  May be overridden by column
        metadata for a specific column. Defaults to 4.
    missing_val : str
        A value that will replace any cell data having a value of None.  Defaults to ''.
    max_width : int or None
        If not None, specifies the maximum width, in characters, allowed for the table.
        If the table cannot meet the max width requirement by resizing columns, the max_width
        is ignored.

    Attributes
    ----------
    missing_val : str
        String to replace any data values of None.
    max_width : int or None
        If not None, specifies the maximum allowable width, in characters, of the table.
    _ncols : int
        Number of columns.
    _raw_rows : iter of iters
        Table row data from the caller, possibly converted to list of lists if caller passed in
        dict or iter of dicts.
    _rows : list of lists
        Table row data after initial formatting.
    _column_meta : dict
        Metadata for each column, keyed by column index, starting at 0.
    _data_widths : list of int
        Width of widest data cell in each column.
    _header_widths : list of int
        Width of each column header.
    _default_formats : dict
        Dict mapping each column type to its default format string.
    """

    allowed_col_meta = {'header', 'align', 'header_align', 'width', 'format',
                        'max_width', 'min_width', 'fixed_width'}

    def __init__(self, rows, headers=None, column_meta=None, precision=4, missing_val='',
                 max_width=None):
        """
        Initialize all attributes.
        """
        if headers in ('keys', 'firstrow'):
            rows, headers = self._to_rows(rows, headers)
        elif isinstance(headers, str):
            raise RuntimeError(f"If 'headers' is a string, it must be one of ['keys', 'firstrow'].")

        self._raw_rows = []
        for row in rows:
            self._raw_rows.append(list(row))
        self._ncols = _num_cols(self._raw_rows)
        self._rows = None  # rows after initial formatting (total width not set)
        self._column_meta = {}
        self._data_widths = None  # width of data in each cell before a uniform column width is set
        self._header_widths = None  # width of headers before a uniform column width is set
        self.missing_val = missing_val
        self.max_width = max_width

        # these are the default format strings for the first formatting stage,
        # before the column width is set
        self._default_formats = {
            'real': f"{{:.{precision}}}",
            'int': "{}",
            'bool': "{}",
            'other': "{}",
        }

        # for convenience, allow a user to specify header strings without putting them
        # inside a metadata dict
        if headers is not None:
            headers = list(headers)
            hlen = len(headers)
            if hlen != self._ncols:
                raise RuntimeError("Number of headers and number of data columns must match, but "
                                   f"{hlen} != {self._ncols}.")

            for i, h in enumerate(headers):
                if not isinstance(h, str):
                    h = str(h)
                self.update_column_meta(i, header=h)

        if column_meta is not None:
            column_meta = list(column_meta)
            clen = len(column_meta)
            if clen != self._ncols:
                raise RuntimeError("Number of column metadata dicts and number of data columns "
                                   f"must match, but {clen} != {self._ncols}.")

            for i, meta in enumerate(column_meta):
                self.update_column_meta(i, **meta)

        if headers is not None and column_meta is not None and hlen != clen:
            raise RuntimeError("Number of headers and number of column metadata dicts must match "
                               f"if both are provided, but {hlen} != {clen}.")

    def display(self, outfile=None):
        """
        Display this table.

        Parameters
        ----------
        outfile : str or None
            If None, print this table to stdout, else write it to the named file.
        """
        raise NotImplementedError("The display method is not defined for class "
                                  f"'{type(self).__name__}'.")

    def _to_rows(self, rows, headers):
        """
        Convert dict or iter of dicts into expected row and header data format.

        Parameters
        ----------
        rows : dict or iter of dicts
            Table cell and header data.
        headers : str
            Either 'keys' or 'firstrow'.

        Returns
        -------
        list of lists
            Table data cells.
        list of str
            Table headers.
        """
        new_rows = []

        if headers == 'firstrow':

            for i, row in enumerate(rows):
                if i == 0:
                    headers = list(row)
                else:
                    new_rows.append(list(row))

            return new_rows, headers

        headers = []

        if isinstance(rows, dict):
            # each value is a column, so re-arrange to be row major, and allow columns of
            # unequal length (to be compatible with tabulate).
            headers = list(rows.keys())

            # First, get max column length
            maxcol = 0
            for col in rows.values():
                clen = len(list(col))
                if clen > maxcol:
                    maxcol = clen

            for col in rows.values():
                if not new_rows:
                    new_rows = [[] for i in range(maxcol)]

                for row, cell in zip_longest(new_rows, col, fillvalue=''):
                    row.append(cell)
        else:
            # handle case where rows is an iter of dicts
            try:
                for row in rows:
                    headers = list(row.keys())
                    break
            except AttributeError:
                raise AttributeError("Since headers == 'keys', table builder was expecting an iter "
                                     f"of dicts, but got an iter of {type(row).__name__}.")
            for row in rows:
                new_rows.append(list(row.values()))

        return new_rows, headers

    def sorted_meta(self):
        """
        Return the column metadata sorted in ascending column order.

        Returns
        -------
        list
            The sorted column metadata.
        """
        return [m for _, m in sorted(self._column_meta.items(), key=lambda x: x[0])]

    def _get_formatted_rows(self):
        """
        Get table rows with cells after initial formatting, before final width setting.

        Returns
        -------
        list
            The list of table rows where each cell has been formatted.
        """
        if self._rows is None:
            self._update_col_meta_from_rows()
            sorted_cols = self.sorted_meta()
            self._rows = []
            unequal = False
            maxcols = 0
            for row in self._raw_rows:
                if self.missing_val is not None:
                    row = [self.missing_val if v is None else v for v in row]

                if self._rows and len(row) != len(self._rows[-1]):
                    unequal = True

                if len(row) > maxcols:
                    maxcols = len(row)

                self._rows.append([self._format_cell(meta, cell)
                                   for meta, cell in zip(sorted_cols, row)])

            if unequal:
                # make all rows have same number of cells
                for row in self._rows:
                    if len(row) < maxcols:
                        row.extend([self.missing_val] * (maxcols - len(row)))

        return self._rows

    def _cell_width(self, cell):
        """
        Return the width of the cell.

        Returns
        -------
        int
            The width of the cell.
        """
        return len(cell)

    def _get_cell_types(self):
        """
        Yield the type of cells each column contains.

        If columns have mixed types, yield 'other'.

        Yields
        ------
        str
            The type of the current column.
        """
        types = []
        for row in self._raw_rows:
            if not types:
                types = [set() for r in row] if row is not None else []

            for i, cell in enumerate(row):
                if isinstance(cell, Number):
                    if isinstance(cell, bool):
                        types[i].add('bool')
                    elif isinstance(cell, Integral):
                        types[i].add('int')
                    else:
                        types[i].add('real')
                elif cell is None:
                    pass  # don't add to types if cell is None
                else:
                    types[i].add('other')

        for tset in types:
            if len(tset) > 1 or len(tset) == 0:
                yield 'other'  # mixed type column (or all Nones), so just use "{}" format
            else:
                yield tset.pop()

    def update_column_meta(self, col_idx, **options):
        r"""
        Update metadata for the column at the specified index (starting at index 0).

        Parameters
        ----------
        col_idx : int
            The index of the column, starting at 0.
        **options : dict
            The metadata dict will be updated with these options.
        """
        if col_idx < 0:  # allow negative indices
            col_idx = self._ncols + col_idx

        if col_idx < 0 or col_idx >= self._ncols:
            raise IndexError(f"Index '{col_idx}' is not a valid table column index for a table with"
                             f" {self._ncols} columns.")

        if col_idx not in self._column_meta:
            self._column_meta[col_idx] = {}

        meta = self._column_meta[col_idx]
        for name, val in options.items():
            meta[name] = val

    def _set_widths(self, force_set_max=False):
        """
        Set data and header widths to their final values.

        Parameters
        ----------
        force_set_max : bool
            If True, compute the max column widths even if the table max_width limit is violated.
        """
        if self._data_widths is not None:
            return  # widths already computed

        rows = self._get_formatted_rows()

        if self._rows and len(self._rows[0]) != self._ncols:
            raise RuntimeError(f"Number of row entries ({len(self._rows[0])}) must match number of "
                               f"columns ({self._ncols}) in TableBuilder.")

        self._data_widths = [0] * self._ncols
        self._header_widths = [0] * self._ncols

        for row in rows:
            for i, cell in enumerate(row):
                wid = self._cell_width(cell)
                if wid > self._data_widths[i]:
                    self._data_widths[i] = wid

        # set widths and min widths
        for i, meta in enumerate(self.sorted_meta()):
            self._header_widths[i] = 0 if meta['header'] is None else len(meta['header'])

        self._set_min_widths()
        self._set_max_column_widths(force_set_max)

    def _format_cell(self, meta, cell):
        """
        Apply the initial formatting (before final width setting) to a cell.

        Parameters
        ----------
        meta : dict
            Column metadata dict.
        cell : object
            Data contained in a table cell.
        """
        if cell is None:
            return ''
        try:
            return meta['format'].format(cell)
        except Exception:
            return f'{cell}'

    def _update_col_meta_from_rows(self):
        """
        Fill in missing column metadata based on the data types of column contents.
        """
        for i, col_type in enumerate(self._get_cell_types()):
            align = _default_align[col_type]

            meta = {
                'header': None,
                'format': self._default_formats[col_type],
                'align': align,
                'header_align': align,
                'max_width': None,
                'col_type': col_type
            }

            if i in self._column_meta:
                meta.update(self._column_meta[i])

            self._column_meta[i] = meta

    def needs_wrap(self):
        """
        Return True if the width of the table or any column exceeds the its specified max_width.

        Returns
        -------
        bool
            Indicates whether or not this table requires word wrapping to fit the maximum width.
        """
        needs_wrap = self.max_width is not None and self.max_width < self._get_total_width()
        if not needs_wrap:
            for meta in self._column_meta.values():
                if meta['max_width'] is not None:
                    return True
        return needs_wrap

    def _set_min_widths(self):
        """
        Set minimum width for columns.
        """
        for meta, wcell in zip(self.sorted_meta(), self._data_widths):
            header = meta['header']

            # check if header is splittable into words, and if so, allow for min_width using a
            # split header if min_width will be > min data width.
            if header is not None and header.strip():
                longest_part = max(len(word) for word in header.strip().split())
            else:
                longest_part = 0

            if meta['col_type'] == 'other':  # strings
                meta['min_width'] = max(10, longest_part)
            else:
                meta['min_width'] = max(wcell, longest_part)

    def _set_max_column_widths(self, force_set_max=False):
        """
        Set the maximum allowable column widths based on the table max_width.

        Parameters
        ----------
        force_set_max : bool
            If True, compute the max column widths even if the table max_width limit is violated.
        """
        # check for case where total table width is specified and we have to set max_width on
        # column(s) as a result
        if self.max_width is not None and self.max_width < self._get_total_width():
            winfo = [[i, w, meta['min_width']]
                     for (i, meta), w in zip(enumerate(self.sorted_meta()),
                                             self._get_column_widths())
                     if not meta.get('fixed_width')]

            fixed_width = self._get_total_width() - sum([w for _, w, _ in winfo])
            allowed_width = self.max_width - fixed_width

            # subtract 1 from the widest column until we meet the total max_width requirement,
            # or get as close as we can without violating a minimum allowed width.
            while sum([w for _, w, _ in winfo]) > allowed_width:
                for info in sorted(winfo, key=lambda x: x[1], reverse=True):
                    _, width, min_width = info
                    if width - 1 >= min_width:
                        info[1] -= 1
                        break
                else:
                    break

            if force_set_max or sum([w for _, w, _ in winfo]) <= allowed_width:
                for i, w, _ in winfo:
                    self._column_meta[i]['max_width'] = w

    def _get_column_widths(self):
        """
        Return the width of each column.

        Returns
        -------
        list
            The column widths.
        """
        return [max(wd, wh) for wd, wh in zip(self._data_widths, self._header_widths)]

    def write(self, outfile=None):
        """
        Write this table to the given output file.

        Parameters
        ----------
        outfile : str or None
            The output file.  If None, assume table should be sent to stdout.

        Returns
        -------
        str or None
            The output file name or None if written to stdout.
        """
        if outfile is None:
            sys.stdout.write(str(self))
        else:
            with open(outfile, 'w') as f:
                f.write(str(self))
            return outfile


@dataclass(frozen=True)
class Line:
    """
    Information about a line in the table.

    Parameters
    ----------
    left : str
        Left border string.
    sep : str
        Column separator.
    right : str
        Right border string.
    hline : str
        Horizontal line string.

    Attributes
    ----------
    left : str
        Left border string.
    sep : str
        Column separator.
    right : str
        Right border string.
    hline : str
        Horizontal line string.
    """

    left: str = ''
    sep: str = ''
    right: str = ''
    hline: str = ''

    def get_border_line(self, widths):
        """
        Return a border line given the column widths.

        Parameters
        ----------
        widths : list of int
            Column widths.

        Returns
        -------
        str
            The border line.
        """
        line = self.sep.join([(self.hline * w)[:w] for w in widths])
        return ''.join((self.left, line, self.right))

    def get_data_line(self, cells):
        """
        Return a table line containing the given cells.

        Parameters
        ----------
        cells : list of str
            Contents of table columns for the current row.

        Returns
        -------
        str
            The table line containing the cells, with separators and left and right borders.
        """
        return ''.join((self.left, self.sep.join(cells), self.right))


class TextTableBuilder(TableBuilder):
    r"""
    Base class for all text-based table builders.

    Parameters
    ----------
    rows : iter of iters
        Data used to fill table cells.
    top_border : Line
        Top border info.
    header_bottom_border : Line
        Header bottom border info.
    bottom_border : Line
        Bottom border info.
    header_line : Line
        Header line info.
    data_row_line : Line
        Data row line info.
    row_separator : Line or None
        If not None, info for lines between data rows.
    **kwargs : dict
        Keyword args for the base class.

    Attributes
    ----------
    top_border : Line
        Top border info.
    header_bottom_border : Line
        Header bottom border info.
    bottom_border : Line
        Bottom border info.
    header_line : Line
        Header line info.
    data_row_line : Line
        Data row line info.
    row_separator : Line or None
        If not None, info for lines between data rows.
    """

    def __init__(self, rows,
                 top_border=Line('| ', '---', ' |', '-'),
                 header_bottom_border=Line('| ', ' | ', ' |', '-'),
                 bottom_border=Line('| ', '---', ' |', '-'),
                 header_line=Line('| ', ' | ', ' |'),
                 data_row_line=Line('| ', ' | ', ' |'),
                 row_separator=None,
                 **kwargs):
        """
        Initialize all attributes.
        """
        super().__init__(rows, **kwargs)

        self.top_border = top_border
        self.header_line = header_line
        self.header_bottom_border = header_bottom_border
        self.bottom_border = bottom_border
        self.data_row_line = data_row_line
        self.row_separator = row_separator

    def _get_total_width(self):
        """
        Return the total table width in characters.
        """
        tot = sum(self._get_column_widths())
        tot += (len(self._column_meta) - 1) * len(self.top_border.sep)
        tot += len(self.top_border.left) + len(self.top_border.right)
        return tot

    def _get_fixed_width_cell(self, col_meta, cell, width, align_name):
        """
        Return a string of the specified width and alignment for the given cell.

        Parameters
        ----------
        col_meta : dict
            Metadata for the current column.
        cell : object
            The cell data.
        width : int
            The desired width of the returned cell string.
        align_name : str
            The name of the alignment metadata, either 'align' or 'header_align', used to retrieve
            the corresponding alignment symbol needed for string formatting.

        Returns
        -------
        str
            A string of the specified width containing the cell data, aligned as specified.
        """
        align = col_meta.get(align_name, 'left')
        try:
            sym = _align2symbol[align]
        except KeyError:
            raise KeyError(f"Expected one of ['left', 'right', 'center'] for '{align_name}' "
                           f"metadata, but got '{align}'.")
        return f"{cell:{sym}{width}}"

    def get_lengthened_columns(self, sorted_cols, header_cells, widths):
        """
        Yield as many rows of cells as needed to allow for multi-line cells due to word wrapping.

        Parameters
        ----------
        sorted_cols : list of dict
            List of sorted column metadata.
        header_cells : list of str
            List of header strings.
        widths : list of int
            Column widths.

        Yields
        ------
        list
            Each row after expanding due to word wrapping.
        """
        cell_lists = []
        for meta, cell, wid in zip(sorted_cols, header_cells, widths):
            maxwid = meta['max_width']
            if maxwid is not None and maxwid < len(cell):
                lines = textwrap.wrap(cell, maxwid)
                wid = maxwid
            elif '\n' in cell:
                lines = cell.split('\n')
            else:
                cell_lists.append([cell])
                continue

            sym = _align2symbol[meta['header_align']]
            if sym == '^':  # center
                lines = [line.strip() for line in lines]
            # ensure all cells have same width in this column
            cell_lists.append([f"{line:{sym}{wid}}" for line in lines])

        # now find longest column
        maxlen = max([len(lst) for lst in cell_lists])
        for r in range(maxlen):
            cells = []
            for clist in cell_lists:
                if len(clist) > r:
                    cells.append(clist[r])
                else:
                    cells.append(' ' * len(clist[0]))
            yield cells

    def _stringified_header_iter(self):
        """
        Yield one or more rows of header cells.

        Yields
        ------
        list
            Each header row after expanding due to word wrapping.
        """
        header_cells = [None] * self._ncols
        self._set_widths()

        widths = self._get_column_widths()

        sorted_cols = self.sorted_meta()

        for i, meta in enumerate(sorted_cols):
            header = meta['header']
            if header is not None:
                header_cells[i] = self._get_fixed_width_cell(meta, header, widths[i],
                                                             'header_align')

        yield from self.get_lengthened_columns(sorted_cols, header_cells, widths)

    def _stringified_row_iter(self):
        """
        Yield rows of data cells, allowing for multi-line rows due to word wrapping.

        The cells are all strings with the same width as their column.

        Yields
        ------
        list
            List of cells for the current row.
        """
        self._set_widths()
        widths = self._get_column_widths()
        row_cells = [None] * len(self._column_meta)

        sorted_cols = self.sorted_meta()

        for row in self._get_formatted_rows():
            for i, meta in enumerate(sorted_cols):
                row_cells[i] = self._get_fixed_width_cell(meta, row[i], widths[i], 'align')

            cell_lists = []
            for meta, cell, wid in zip(sorted_cols, row_cells, widths):
                maxwid = meta['max_width']
                if maxwid is not None and maxwid < len(cell):
                    lines = textwrap.wrap(cell, maxwid)
                    if not lines:  # empty cell or whitespace
                        lines = ['']
                    wid = maxwid
                elif '\n' in cell:
                    lines = cell.split('\n')
                else:
                    cell_lists.append([cell])
                    continue

                sym = _align2symbol[meta['align']]
                if sym == '^':  # center
                    lines = [line.strip() for line in lines]
                # ensure all cells have same width in this column
                cell_lists.append([f"{line:{sym}{wid}}" for line in lines])

            # now find longest column
            maxlen = max([len(lst) for lst in cell_lists])
            cell_list_group = []
            for r in range(maxlen):
                cells = []
                for clist in cell_lists:
                    if len(clist) > r:
                        cells.append(clist[r])
                    else:
                        w = len(clist[0]) if clist else 0
                        cells.append(' ' * w)
                cell_list_group.append(cells)

            yield cell_list_group

    def get_top_border(self, widths):
        """
        Return the top border string for this table.

        Parameters
        ----------
        widths : list of int
            Column widths.

        Returns
        -------
        str
            The top border string.
        """
        return self.top_border.get_border_line(widths)

    def get_header_bottom_border(self, widths):
        """
        Return the header bottom border string for this table.

        Parameters
        ----------
        widths : list of int
            Column widths.

        Returns
        -------
        str
            The header bottom border string.
        """
        return self.header_bottom_border.get_border_line(widths)

    def get_bottom_border(self, widths):
        """
        Return the bottom border string for this table.

        Parameters
        ----------
        widths : list of int
            List of column widths.

        Returns
        -------
        str
            The bottom border string.
        """
        return self.bottom_border.get_border_line(widths)

    def __str__(self):
        """
        Return this table to a string.

        Returns
        -------
        str
            This table as a string.
        """
        header_lines = []
        data_lines = []
        sep = self.data_row_line.sep
        row_cells = None
        row_list = list(self._stringified_row_iter())
        if row_list:
            widths = [len(c) for c in row_list[0][0]]
        else:
            widths = []

        for i, cell_list_group in enumerate(row_list):
            if i > 0 and self.row_separator:
                data_lines.append(self.row_separator.get_border_line(widths))
            for row_cells in cell_list_group:
                data_lines.append(self.data_row_line.get_data_line(row_cells))

        if self.bottom_border:
            data_lines.append(self.get_bottom_border(widths))

        if row_cells is not None and self.top_border:
            header_lines.append(self.get_top_border(widths))

        if sum(self._header_widths) > 0:
            for header_cells in self._stringified_header_iter():
                header_lines.append(self.header_line.get_data_line(header_cells))

            if self.header_bottom_border:
                header_lines.append(self.get_header_bottom_border(widths))

        return '\n'.join(chain(header_lines, data_lines))

    def display(self, outfile=None):
        """
        Display this table.

        Parameters
        ----------
        outfile : str or None
            If None, print this table to stdout, else write it to the named file.

        Returns
        -------
        str or None
            The name of the file where the table was written, or None if it was written to stdout.
        """
        if outfile is None:
            sys.stdout.write(str(self))
            sys.stdout.write('\n')
            return

        with open(outfile, 'w') as f:
            f.write(str(self))
            f.write('\n')

        return outfile


class GithubTableBuilder(TextTableBuilder):
    r"""
    Class that generates a table in Github markdown format.

    Parameters
    ----------
    rows : iter of iters
        Data used to fill table cells.
    **kwargs : dict
        Keyword args for the base class.
    """

    def __init__(self, rows, **kwargs):
        """
        Initialize all attributes.
        """
        super().__init__(rows,
                         top_border=None,
                         header_bottom_border=Line('| ', ' | ', ' |', '-'),
                         bottom_border=None,
                         header_line=Line('| ', ' | ', ' |'),
                         data_row_line=Line('| ', ' | ', ' |'), **kwargs)

    def _repr_markdown_(self):
        return str(self)

    def get_header_bottom_border(self, widths):
        """
        Return the header bottom border string.

        Parameters
        ----------
        widths : list of int
            Column widths.

        Returns
        -------
        str
            The header bottom border string.
        """
        parts = []
        for i, width in enumerate(widths):
            meta = self._column_meta[i]
            align = meta['align']
            left = right = ''
            size = width
            if align == 'left':
                left = ':'
                size -= 1
            elif align == 'right':
                right = ':'
                size -= 1
            else:  # center
                left = right = ':'
                size -= 2
            parts.append(left + (self.header_bottom_border.hline * size) + right)

        return ''.join((self.header_bottom_border.left, self.header_bottom_border.sep.join(parts),
                        self.header_bottom_border.right))

    def needs_wrap(self):
        """
        Return False.

        This table does not allow word wrapping.

        Returns
        -------
        False
            No wrapping will be used.
        """
        return False  # github tables seem to have no support for text wrapping in columns


def _to_inline_style(dct):
    """
    For the given dict, return an inline HTML style string.

    Parameters
    ----------
    dct : dict
        The dict containing style parameters and their values.

    Returns
    -------
    str
        The inline style string.
    """
    parts = ' '.join([f"{name}: {val};" for name, val in dct.items()])
    if parts:
        return f' style="{parts}"'
    return ''


_default_html_table = 'table.html'


class HTMLTableBuilder(TableBuilder):
    r"""
    Class that generates a table in plain HTML format.

    Parameters
    ----------
    rows : iter of iters
        Data used to fill table cells.
    html_id : str or None
        If not None, the HTML id for the <table> block.
    title : str or None
        If not None, the title appearing above the table on the web page.
    center : bool
        If True, center the table on the page.
    style : dict or None
        If not None, a dict mapping table style parameters to their values.
    safe : bool
        If True (the default), html escape text in the cells.
    **kwargs : dict
        Keyword args for the base class.

    Attributes
    ----------
    _html_id : str or None
        If not None, this is the html id of the table block.
    _title : str or None
        If not None, this is the title that will appear on the web page above the table.
    _safe : bool
        If True (the default), html escape text in the cells.
    _data_style : dict
        Contains style metadata for <td> blocks.
    _header_style : dict
        Contains style metadata for <th> blocks.
    _style : dict or None
        Contains style metadata for the table.
    """

    allowed_col_meta = TableBuilder.allowed_col_meta.union({
        'header_style',
        'style',
    })

    def __init__(self, rows, html_id=None, title='', center=False, style=None, safe=True, **kwargs):
        """
        Initialize all attributes.
        """
        super().__init__(rows, **kwargs)
        self._data_style = {
            'border': '1px solid #999',
            'border-collapse': 'collapse',
            'padding': '5px',
        }
        self._header_style = {
            'border': '1px solid #999',
            'border-collapse': 'collapse',
            'padding': '5px',
            'background-color': '#E9E9E9',
        }
        tstyle = {
            'border': '1px solid #999',
            'border-collapse': 'collapse',
        }
        if center:
            tstyle['margin'] = 'auto'

        if style is not None:
            tstyle.update(style)

        self._style = tstyle
        self._html_id = html_id
        self._title = title
        self._safe = safe

    def _repr_html_(self):
        return str(self)

    def _stringified_row_iter(self):
        """
        Yield rows of string data cells whose width matches their column width.

        Yields
        ------
        list
            List of cells for the current row.
        """
        for row in self._get_formatted_rows():
            yield row

    def _assemble(self):
        rlines = []

        def _escape(s):
            if s is None:
                return ''
            elif self._safe:
                return escape(s)
            return s

        for irow, row_cells in enumerate(self._stringified_row_iter()):
            row_style = {'background-color': '#F3F3F3' if irow % 2 else 'ghostwhite'}
            parts = [f"   <tr{_to_inline_style(row_style)}>"]
            for cell, meta in zip(row_cells, self.sorted_meta()):
                style = self._data_style.copy()
                style['text-align'] = meta['align']
                parts.append(f'<td{_to_inline_style(style)}>{_escape(cell)}</td>')
            parts.append("</tr>")
            rlines.append(''.join(parts))

        lines = []
        html_id = f' id="{self._html_id}"' if self._html_id else ''
        lines.append(f'<table{html_id}{_to_inline_style(self._style)}>')

        # we do the header out-of-order with the rows here because we need to
        # query the rows first to determine column data type to fill in missing
        # parts of the column metadata.

        # First, make sure we actually have a header
        has_header = False
        for meta in self.sorted_meta():
            if meta['header'] is not None:
                has_header = True
                break

        if has_header:
            hparts = ["    <tr>"]
            for meta in self.sorted_meta():
                style = self._header_style.copy()
                style['text-align'] = meta['header_align']
                if 'header_style' in meta and meta['header_style']:
                    style.update(meta['header_style'])
                header_style = _to_inline_style(style)
                hparts.append(f"<th{header_style}>{_escape(meta['header'])}</th>")
            hparts.append("</tr>")
            lines.append(''.join(hparts))

        lines.extend(rlines)  # now add the rows back in

        lines.append("</table>")

        return '\n'.join(lines)

    def __str__(self):
        """
        Return a string representation of the Table.
        """
        return textwrap.dedent("""
            <!DOCTYPE html>
            <html lang="en">
            <head>
                <style>
                    h2 {{
                        text-align: center;
                    }}
                </style>
            </head>
            <body>
                <h2>{}</h2>
                {}
            </body>
            </html>
            """).format(self._title, textwrap.indent(self._assemble(), '    '))

    def write(self, outfile=_default_html_table):
        """
        Write this table to the given output file.

        If outfile is not given, write this table to 'table.html' in the current
        directory.

        Parameters
        ----------
        outfile : str
            The output file where the HTML will be written.

        Returns
        -------
        str
            The name of the file where the table was written.
        """
        with open(outfile, 'w', encoding='utf-8') as f:
            f.write(str(self))
        return outfile

    def display(self, outfile=None):
        """
        Display this table, either in a notebook or a browser.

        This table will also be written to outfile.  If outfile is not supplied, the table
        will be written to 'table.html' in the current directory.

        Parameters
        ----------
        outfile : str or None
            Table will be written to this file.
            If None, write this table to '{_default_html_table}'.

        Returns
        -------
        str
            The name of the file where the table was written.
        """
        if outfile is None:
            outfile = _default_html_table

        self.write(outfile)

        if notebook:
            if not colab:
                height = min(_big_table_height, len(self._rows) * _row_height_px)
                display(IFrame(src=outfile, width="100%", height=height))
            else:
                display(HTML(outfile))
        else:
            # open it up in the browser
            from openmdao.utils.webview import webview
            webview(outfile)

        return outfile


_tabulator_typemeta = {
    'bool': {
        'align': 'center',
        'formatter': 'tickCross',
        'formatterParams': {'crossElement': False},
        'titleFormatter': 'textarea',
        'filter': 'tickCross',
        'headerFilterParams': {'tristate': True},
        'sorter': 'string',
    },
    'int': {
        'align': 'right',
        'formatter': 'plaintext',
        'titleFormatter': 'textarea',
        'filter': False,
        'sorter': 'number',
    },
    'real': {
        'align': 'right',
        'formatter': 'plaintext',
        'titleFormatter': 'textarea',
        'filter': False,
        'sorter': 'number',
    },
    'other': {
        'align': 'left',
        'formatter': 'textarea',
        'titleFormatter': 'textarea',
        'filter': 'input',
        'sorter': 'string',
    }
}

_default_tabulator_file = 'tabulator_table.html'


class TabulatorJSBuilder(TableBuilder):
    r"""
    Class that generates an interactive table using Tabulator.js.

    Parameters
    ----------
    rows : iter of iters
        Data used to fill table cells.
    html_id : str or None
        If not None, the HTML id for the <table> block.
    title : str or None
        If not None, the title appearing above the table on the web page.
    filter : bool
        If True, include filter fields in the column headers where it makes sense.
    sort : bool
        If True, add sorting to column headers.
    center : bool
        If True, center the table on the page.
    table_meta : dict or None
        If not None, a dict of Tabulator.js metadata names mapped to their values.
    **kwargs : dict
        Keyword args for the base class.

    Attributes
    ----------
    _html_id : str or None
        If not None, this is the html id of the table block.
    _title : str or None
        If not None, this is the title that will appear on the web page above the table.
    _filter : bool
        If True, include filter fields in the column headers where it makes sense.
    _sort : bool
        If True, add sorting to column headers.
    _center : bool
        If True, center the table on the page.
    _table_meta : dict
        Metadata for the table.
    font_size : int
        The font size used by the table.
    """

    allowed_col_meta = TableBuilder.allowed_col_meta.union({
        'filter',
        'header_align',
        'sorter',
        'formatter',
        'formatterParams',
        'labelField',
        'target',
    })

    def __init__(self, rows, html_id='tabul-table', title='', filter=True, sort=True, center=False,
                 table_meta=None, **kwargs):
        """
        Initialize all attributes.
        """
        super().__init__(rows, **kwargs)
        self._html_id = html_id if html_id.startswith('#') else '#' + html_id
        self._title = title
        self._filter = filter
        self._sort = sort
        self._center = center

        self._table_meta = {
            'autoResize': True,
        }

        if table_meta is not None:
            self._table_meta.update(table_meta)

        self.font_size = 14

    def _cell_width(self, cell):
        """
        Return given cell's width in characters.

        This table's bool cells are left as bools instead of converted to strings in order to
        make the Tabulator.js table filtering work properly for bool columns, so this method
        will return a width for bool cells.

        Parameters
        ----------
        cell : str or bool
            The contents of the current table cell.

        Returns
        -------
        int
            The width of the cell.
        """
        # special handling for bool cells
        if isinstance(cell, bool):
            return 5
        return len(cell)

    def _format_cell(self, meta, cell):
        """
        Return the string formatted form of the cell, but leave bool cells unmodified.

        Parameters
        ----------
        meta : dict
            The metadata for the current column.
        cell : str or bool
            The current table cell.

        Returns
        -------
        str or bool
            The formatted cell if it's a str, else the unmodified cell.
        """
        if isinstance(cell, bool):
            return cell
        return super()._format_cell(meta, cell)

    def _get_total_width(self):
        """
        Return the total table width in characters.

        Returns
        -------
        int
            The total table width.
        """
        return sum(self._get_column_widths())

    def _stringified_row_iter(self):
        """
        Yield rows of string data cells.

        Yields
        ------
        list
            List of cells for the current row.
        """
        self._set_widths(force_set_max=True)
        for row in self._get_formatted_rows():
            yield row

    def _update_col_meta_from_rows(self):
        """
        Fill in missing column metadata based on the data types of column contents.
        """
        for i, col_type in enumerate(self._get_cell_types()):
            meta = _tabulator_typemeta[col_type].copy()
            meta['format'] = self._default_formats[col_type]
            meta['header_align'] = meta['align']
            meta['max_width'] = None
            meta['col_type'] = col_type
            meta['header'] = None

            if i in self._column_meta:
                meta.update(self._column_meta[i])

            self._column_meta[i] = meta

    def _get_table_data(self):
        """
        Return table data in a format that can be converted to json for Tabulator.js to use.

        Returns
        -------
        dict
            A dict to be converted to json and provided to Tabulator.js.
        """
        rows = []
        idx = 1  # unique ID for use by Tabulator

        for row_cells in self._stringified_row_iter():
            dct = {'id': idx}
            for i, cell in enumerate(row_cells):
                dct[f'c{i}'] = cell
            rows.append(dct)
            idx += 1

        if 'layout' not in self._table_meta:
            if self.needs_wrap():
                self._table_meta['layout'] = 'fitColumns'
            else:
                self._table_meta['layout'] = 'fitDataTable'

        if self._table_meta['layout'] == 'fitColumns':
            self._setup_fit_columns_layout()

        cols = []
        for i, meta in enumerate(self.sorted_meta()):
            cmeta = {
                'field': f'c{i}',
                # can't eliminate headers from tabulator.js tables, so just set to ' ' if None.
                # Setting to '' results in the header containing '&nbsp' which doesn't look great.
                'title': ' ' if meta['header'] is None else meta['header'],
                'hozAlign': meta['align'],
                'headerHozAlign': meta['header_align'],
                'headerFilter': meta['filter'] if self._filter else False,
                'sorter': meta['sorter'] if self._sort else False,
                'headerSort': self._sort,
                'formatter': meta['formatter'],  # plaintext, textarea, html, money, image, link,
                                                 # tickCross, traffic, star, progress, color,
                                                 # buttonTick, buttonCross,
                'formatterParams': meta.get('formatterParams'),
                'titleFormatter': meta.get('titleFormatter'),
                'titleFormatterParams': meta.get('titleFormatterParams'),
                'editor': meta.get('editor'),
                'editorParams': meta.get('editorParams'),
                'headerFilterParams': meta.get('headerFilterParams'),
                'widthGrow': meta.get('widthGrow'),
                'widthShrink': meta.get('widthShrink'),
                'tooltip': meta.get('tooltip'),
                'visible': meta.get('visible', True),
            }

            width = meta.get('width')
            if width is not None:
                cmeta['width'] = width
            cols.append(cmeta)

        # for big tables, use virtual DOM for speed (setting height activates it)
        if idx - 1 > 60 and ('height' not in self._table_meta or
                             self._table_meta['height'] is None):
            self._table_meta['height'] = _big_table_height

        self._table_meta['data'] = rows
        self._table_meta['columns'] = cols

        return {
            'id': self._html_id,
            'title': self._title,
            'meta': self._table_meta,
        }

    def _setup_fit_columns_layout(self):
        smeta = self.sorted_meta()
        maxws = [m['max_width'] for m in smeta]
        if None in maxws:
            issue_warning("Can't use 'layout' of 'fitColumns' with columns that have no "
                          "max_width set. Switching to layout of 'fitDataTable'.")
            self._table_meta['layout'] = 'fitDataTable'
            return

        mx = sum(maxws)
        pcts = [int(w / mx * 100) for w in maxws]
        rempcts = 100 - sum(pcts)
        while rempcts > 0:
            for i in range(len(pcts)):
                if rempcts > 0:
                    pcts[i] += 1
                    rempcts -= 1
                else:
                    break

        for pct, meta in zip(pcts, smeta):
            meta['width'] = f"{pct}%"

    def write(self, outfile=_default_tabulator_file):
        """
        Write this table to the given output file.

        If outfile is not given, write this table to 'tabulator_table.html' in the current
        directory.

        Parameters
        ----------
        outfile : str or None
            If not None, the output file where the HTML will be written.

        Returns
        -------
        str
            The output file name.
        """
        outfile = os.path.relpath(outfile)

        with open(outfile, 'w', encoding='utf-8') as f:
            f.write(str(self))

        return outfile

    def __str__(self):
        """
        Return a string representation of the Table.

        Returns
        -------
        str
            This table as a string.
        """
        import openmdao.visualization

        code_dir = os.path.dirname(openmdao.visualization.__file__)
        libs_dir = os.path.join(code_dir, 'common', 'libs')
        style_dir = os.path.join(code_dir, 'common', 'style')

        format_dct = {}
        if self._center:
            format_dct['table_div_style'] = 'class="center"'
        else:
            format_dct['table_div_style'] = ''

        with open(os.path.join(code_dir, 'tables', 'generic_table.template'), "r",
                  encoding='utf-8') as f:
            template = f.read()

        with open(os.path.join(libs_dir, 'tabulator.5.4.4.min.js'), "r", encoding='utf-8') as f:
            tabulator_src = f.read()

        format_dct['tabulator_src'] = tabulator_src

        with open(os.path.join(style_dir, 'tabulator.5.4.4.min.css'), "r", encoding='utf-8') as f:
            tabulator_style = f.read()

        format_dct['tabulator_style'] = tabulator_style

        jsontxt = json.dumps(self._get_table_data())
        format_dct['table_data'] = jsontxt

        return template.format(**format_dct)

    def display(self, outfile=None):
        """
        Display this table, either in a notebook or a browser.

        This table will also be written to outfile.  If outfile is not supplied, the table
        will be written to '{_default_tabulator_file}' in the current directory.

        Parameters
        ----------
        outfile : str or None
            If None, write this table to '{_default_tabulator_file}'.

        Returns
        -------
        str
            The name of the file where the table was written.
        """
        if outfile is None:
            outfile = _default_tabulator_file

        self.write(outfile)
        if notebook:
            if not colab:
                height = min(_big_table_height, len(self._rows) * _row_height_px)
                display(IFrame(src=outfile, width="100%", height=height))
            else:
                display(HTML(outfile))
        else:
            # open it up in the browser
            from openmdao.utils.webview import webview
            webview(outfile)

        return outfile


def generate_table(rows, tablefmt='text', **options):
    r"""
    Return the specified table builder class.

    Parameters
    ----------
    rows : iter of iters
        This specifies the cells in each table row.
    tablefmt : str
        This determines which table builder object will be returned.
    **options : dict
        Named arguments specific to a given table builder class.

    Returns
    -------
    TableBuilder
        A table builder object.
    """
    _text_formats = {
        'rst': {
            'top_border': Line('', '  ', '', '='),
            'header_line': Line('', '  ', ''),
            'header_bottom_border': Line('', '  ', '', '='),
            'bottom_border': Line('', '  ', '', '='),
            'data_row_line': Line('', '  ', '')
        },
        'grid': {
            'top_border': Line('+-', '-+-', '-+', '-'),
            'header_line': Line('| ', ' | ', ' |'),
            'header_bottom_border': Line('+=', '=+=', '=+', '='),
            'bottom_border': Line('+-', '-+-', '-+', '-'),
            'data_row_line': Line('| ', ' | ', ' |'),
            'row_separator': Line('+-', '-+-', '-+', '-')
        },
        'simple_grid': {
            'top_border': Line("┌─", "─┬─", "─┐", "─"),
            'header_line': Line("│ ", " │ ", " │"),
            'header_bottom_border': Line("├─", "─┼─", "─┤", "─"),
            'row_separator': Line("├─", "─┼─", "─┤", "─"),
            'bottom_border': Line("└─", "─┴─", "─┘", "─"),
            'data_row_line': Line("│ ", " │ ", " │"),
        },
        'heavy_grid': {
            'top_border': Line("┏━", "━┳━", "━┓", "━"),
            'header_line': Line("┃ ", " ┃ ", " ┃"),
            'header_bottom_border': Line("┣━", "━╋━", "━┫", "━"),
            'row_separator': Line("┣━", "━╋━", "━┫", "━"),
            'bottom_border': Line("┗━", "━┻━", "━┛", "━"),
            'data_row_line': Line("┃ ", " ┃ ", " ┃"),
        },
        'double_grid': {
            'top_border': Line("╔═", "═╦═", "═╗", "═"),
            'header_line': Line("║ ", " ║ ", " ║"),
            'header_bottom_border': Line("╠═", "═╬═", "═╣", "═"),
            'row_separator': Line("╠═", "═╬═", "═╣", "═"),
            'bottom_border': Line("╚═", "═╩═", "═╝", "═"),
            'data_row_line': Line("║ ", " ║ ", " ║"),
        },
        'box_grid': {
            'top_border': Line("╔═", "═╤═", "═╗", "═"),
            'header_line': Line("║ ", " ┊ ", " ║"),
            'header_bottom_border': Line("╠═", "═╪═", "═╣", "═"),
            'row_separator': Line("╟┈", "┈┿┈", "┈╢", "┈"),
            'bottom_border': Line("╚═", "═╧═", "═╝", "═"),
            'data_row_line': Line("║ ", " ┊ ", " ║"),
        },
    }

    for fmt in list(_text_formats):
        if 'grid' in fmt:
            dct = _text_formats[fmt].copy()
            dct['row_separator'] = None
            _text_formats[fmt.replace('grid', 'outline')] = dct

    _table_types = {
        'text': TextTableBuilder,
        'github': GithubTableBuilder,
        'tabulator': TabulatorJSBuilder,
        'html': HTMLTableBuilder,
    }

    if tablefmt in _text_formats:
        kwargs = _text_formats[tablefmt].copy()
        kwargs['rows'] = rows
        kwargs.update(**options)
        return TextTableBuilder(**kwargs)

    try:
        builder = _table_types[tablefmt]
    except Exception:
        raise KeyError(f"'{tablefmt}' is not a valid type choice for generate_table. Valid "
                       f"choices are: {sorted(_table_types)}.")

    return builder(rows, **options)




import unittest
import gc
from contextlib import contextmanager
from types import FunctionType, MethodType, CoroutineType, GeneratorType, FrameType
from collections import defaultdict
from io import StringIO

import openmdao.api as om
from openmdao.core.tests.test_coloring import run_opt
from openmdao.core.system import System
from openmdao.vectors.vector import Vector
from openmdao.solvers.solver import Solver
from openmdao.core.driver import Driver
from openmdao.jacobians.jacobian import Jacobian
from openmdao.approximation_schemes.approximation_scheme import ApproximationScheme
from openmdao.utils.general_utils import set_pyoptsparse_opt
from openmdao.devtools.memory import check_iter_leaks, list_iter_leaks
from openmdao.utils.testing_utils import use_tempdirs

try:
    import objgraph
except ImportError:
    objgraph = None


OPT, OPTIMIZER = set_pyoptsparse_opt('SNOPT', fallback=True)


def run_opt_wrapper(driver_class, optimizer):
    def _wrapper():
        run_opt(driver_class, 'auto', optimizer=optimizer,
                dynamic_total_coloring=True, partial_coloring=True)
    return _wrapper


@unittest.skipUnless(objgraph is not None, "Test requires objgraph to be installed. (pip install objgraph).")
@use_tempdirs
class LeakTestCase(unittest.TestCase):

    ISOLATED = True

    @unittest.skipIf(OPTIMIZER is None, 'pyoptsparse SLSQP is not installed.')
    def test_leaks_pyoptsparse_slsqp(self):
        lst = check_iter_leaks(4, run_opt_wrapper(om.pyOptSparseDriver, 'SLSQP'))
        if lst:
            msg = StringIO()
            list_iter_leaks(lst, msg)
            self.fail(msg.getvalue())

    @unittest.skipUnless(OPTIMIZER == 'SNOPT', 'pyoptsparse SNOPT is not installed.')
    def test_leaks_pyoptsparse_snopt(self):
        lst = check_iter_leaks(4, run_opt_wrapper(om.pyOptSparseDriver, 'SNOPT'))
        if lst:
            msg = StringIO()
            list_iter_leaks(lst, msg)
            self.fail(msg.getvalue())

    def test_leaks_scipy_slsqp(self):
        lst = check_iter_leaks(4, run_opt_wrapper(om.ScipyOptimizeDriver, 'SLSQP'))
        if lst:
            msg = StringIO()
            list_iter_leaks(lst, msg)
            self.fail(msg.getvalue())


if __name__ == '__main__':
    unittest.main()
